from django.contrib import admin
from django.contrib.contenttypes.admin import GenericTabularInline

from ..models import InstrumentMetric


# Register your models here.
class InstrumentMetricGenericInline(GenericTabularInline):
    model = InstrumentMetric

    ordering = ["-date", "key"]
    ct_field = "basket_content_type"
    ct_fk_field = "basket_id"
    raw_id_fields = ["basket_content_type", "instrument", "parent_metric"]

    fields = [
        "date",
        "key",
        "metrics",
    ]

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("basket_content_type", "instrument")


@admin.register(InstrumentMetric)
class InstrumentMetricModelAdmin(admin.ModelAdmin):
    list_filter = ("key",)
    search_fields = ("instrument__computed_str",)
    list_display = (
        "basket_content_type",
        "basket_id",
        "instrument",
        "date",
        "key",
    )

    autocomplete_fields = [
        "basket_content_type",
        "instrument",
    ]
    ordering = ("-date", "-key")
    raw_id_fields = ["basket_content_type", "instrument", "parent_metric"]



from django.contrib import admin

from wbfdm.admin.instruments import InstrumentModelAdmin as BaseInstrumentModelAdmin
from wbfdm.models.instruments import Instrument

from .metrics import InstrumentMetricGenericInline

admin.site.unregister(Instrument)


@admin.register(Instrument)
class InstrumentModelAdmin(BaseInstrumentModelAdmin):
    inlines = (*BaseInstrumentModelAdmin.inlines, InstrumentMetricGenericInline)



from django.conf import settings
from wbcore.utils.importlib import import_from_dotted_path
from .instruments import (
    InstrumentModelViewSet,
    InstrumentRepresentationViewSet,
    InstrumentTypeRepresentationViewSet,
    ChildrenInstrumentModelViewSet,
)
from .instrument_prices import (
    InstrumentPriceModelViewSet,
    InstrumentPriceInstrumentModelViewSet,
    InstrumentPriceInstrumentStatisticsChartView,
    FinancialStatisticsInstrumentPandasView,
    InstrumentPriceInstrumentDistributionReturnsChartView,
    BestAndWorstReturnsInstrumentPandasView,
)

from .classifications import (
    ClassificationClassificationGroupModelViewSet,
    ClassificationGroupModelViewSet,
    ClassificationGroupRepresentationViewSet,
    ClassificationIcicleChartView,
    ClassificationInstrumentThroughInstrumentModelViewSet,
    ClassificationRepresentationViewSet,
    ClassificationTreeChartView,
    InstrumentClassificationThroughInstrumentModelViewSet,
    InstrumentClassificationThroughModelViewSet,
)
from .financials_analysis import (
    CashFlowAnalysisInstrumentBarChartViewSet,
    CashFlowAnalysisInstrumentTableViewSet,
    EarningsInstrumentChartViewSet,
    FinancialsGraphInstrumentChartViewSet,
    NetDebtAndEbitdaInstrumentChartViewSet,
    ProfitabilityRatiosInstrumentChartViewSet,
    SummaryTableInstrumentChartViewSet,
    ValuationRatiosChartView,
)
from .instrument_lists import (
    InstrumentListModelViewSet,
    InstrumentListRepresentationModelViewSet,
    InstrumentListThroughModelViewSet,
    InstrumentListThroughModelInstrumentListViewSet,
    InstrumentListThroughModelInstrumentViewSet,
)

from .instrument_requests import (
    InstrumentRequestModelViewSet,
    InstrumentRequestRepresentationViewSet,
)
from .instruments_relationships import (
    InstrumentClassificationRelatedInstrumentModelViewSet,
    InstrumentClassificationRelatedInstrumentRepresentationViewSet,
    InstrumentFavoriteGroupModelViewSet,
    InstrumentFavoriteGroupRepresentationViewSet,
    RelatedInstrumentThroughInstrumentModelViewSet,
)


ClassificationModelViewSet = import_from_dotted_path(
    getattr(
        settings,
        "DEFAULT_CLASSIFICATION_MODEL_VIEWSET",
        "wbfdm.viewsets.instruments.classifications.ClassificationModelViewSet",
    )
)
ChildClassificationParentClassificationModelViewSet = import_from_dotted_path(
    getattr(
        settings,
        "DEFAULT_PARENT_CLASSIFICATION_MODEL_VIEWSET",
        "wbfdm.viewsets.instruments.classifications.ChildClassificationParentClassificationModelViewSet",
    )
)
ClassifiedInstrumentModelViewSet = import_from_dotted_path(
    getattr(
        settings,
        "DEFAULT_CLASSIFICATION_INSTRUMENT_MODEL_VIEWSET",
        "wbfdm.viewsets.instruments.instruments_relationships.ClassifiedInstrumentModelViewSet",
    )
)



from datetime import date

from django.db import models

from wbfdm.models.instruments import Instrument


class DataBackendMixin:
    def is_object_valid(self, obj: models.Model) -> bool:
        return super().is_object_valid(obj) and obj.is_active_at_date(date.today())

    def get_provider_id(self, obj: models.Model) -> str:
        return self.client.fetch_org_id(org_name=obj.name, org_urls=[obj.primary_url, *obj.additional_urls])

    def get_default_queryset(self):
        return Instrument.objects.filter(instrument_type__key="private_equity")



from datetime import datetime
from typing import Iterator

from django.conf import settings
from wbcore.contrib.dataloader.dataloaders import Dataloader

from wbfdm.dataloaders.protocols import ESGControversyProtocol
from wbfdm.dataloaders.types import ESGControversyDataDict

from ..client import MSCIClient

CONTROVERY_ASSESSMENT_MAP = {
    "Minor": "MINOR",
    "Moderate": "MODERATE",
    "Severe": "SEVERE",
    "Very Severe": "VERY_SEVERE",
}
CONTROVERSY_STATUS_MAP = {
    "Ongoing": "ONGOING",
    "Partially Concluded": "PARTIALLY_CONCLUDED",
    "Concluded": "CONCLUDED",
}
CONTROVERSY_TYPE_MAP = {
    "Structural": "STRUCTURAL",
    "Non-Structural": "NON_STRUCTURAL",
}

CONTROVERSY_FLAG_MAP = {
    "Green": "GREEN",
    "Yellow": "YELLOW",
    "Orange": "ORANGE",
    "Red": "RED",
    "Unknown": "UNKNOWN",
}


class MSCIESGControversyDataloader(ESGControversyProtocol, Dataloader):
    def esg_controversies(self) -> Iterator[ESGControversyDataDict]:
        msci_client_id = getattr(settings, "MSCI_CLIENT_ID", None)
        msci_client_secret = getattr(settings, "MSCI_CLIENT_SECRET", None)
        if msci_client_id and msci_client_secret:
            client = MSCIClient(msci_client_id, msci_client_secret)
            lookup = {k: v for k, v in self.entities.values_list("dl_parameters__esg_controversies__parameters", "id")}
            factors = [
                "CONTROVERSY_CASE_ID",
                "CONTROVERSY_CASE_HEADLINE",
                "CONTROVERSY_LAST_REVIEWED",
                "CONTROVERSY_CASE_NARRATIVE",
                "CONTROVERSY_CASE_SOURCE",
                "CONTROVERSY_CASE_STATUS",
                "CONTROVERSY_CASE_TYPE",
                "CONTROVERSY_CASE_ASSESSMENT",
                "CONTROVERSY_CASE_COMPANY_RESPONSE",
                "CONTROVERSY_CASE_DATE_INITIATED",
                "CONTROVERSY_CASE_FLAG",
            ]
            for row in client.controversies(list(lookup.keys()), factors):
                isin = row.get("CLIENT_IDENTIFIER", "")
                instrument_id = lookup.get(isin, "")

                headline = row.get("CONTROVERSY_CASE_HEADLINE", "")
                narrative = row.get("CONTROVERSY_CASE_NARRATIVE", "")
                reviewed = row.get("CONTROVERSY_LAST_REVIEWED", None)
                initiated = row.get("CONTROVERSY_CASE_DATE_INITIATED", None)
                controversy_id = row.get("CONTROVERSY_CASE_ID", f"{isin}-{hash(headline)}")
                if controversy_id and narrative and headline and isin and instrument_id:
                    yield ESGControversyDataDict(
                        id=controversy_id,
                        instrument_id=instrument_id,
                        headline=headline,
                        narrative=narrative,
                        source=row.get("CONTROVERSY_CASE_SOURCE", ""),
                        review=datetime.strptime(reviewed, "%Y%m%d").date() if reviewed else None,
                        initiated=datetime.strptime(initiated, "%Y%m%d").date() if initiated else None,
                        response=row.get("CONTROVERSY_CASE_COMPANY_RESPONSE", ""),
                        assessment=CONTROVERY_ASSESSMENT_MAP.get(
                            row.get("CONTROVERSY_CASE_ASSESSMENT", "Minor"), None
                        ),
                        status=CONTROVERSY_STATUS_MAP.get(row.get("CONTROVERSY_CASE_STATUS", "Ongoing"), None),
                        type=CONTROVERSY_TYPE_MAP.get(row.get("CONTROVERSY_CASE_TYPE", "Structural"), None),
                        flag=CONTROVERSY_FLAG_MAP.get(row.get("CONTROVERSY_CASE_FLAG", "Green"), None),
                    )



# Generated by Django 5.0.6 on 2024-07-03 14:02

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("currency", "0001_initial"),
        ("geography", "0001_initial"),
        ("io", "0007_alter_exportsource_query_params"),
        ("tags", "0001_initial"),
        ("wbfdm", "0021_delete_instrumentdailystatistics"),
    ]

    operations = [
        migrations.AddField(
            model_name="instrument",
            name="cusip",
            field=models.CharField(blank=True, help_text="CUSIP", max_length=255, null=True, verbose_name="CUSIP"),
        ),
        migrations.AddField(
            model_name="option",
            name="open_interest_20d",
            field=models.FloatField(
                blank=True, help_text="Option Open Interest (20D)", null=True, verbose_name="Open Interest 20D"
            ),
        ),
        migrations.AddField(
            model_name="option",
            name="open_interest_50d",
            field=models.FloatField(
                blank=True, help_text="Option Open Interest (50D)", null=True, verbose_name="Open Interest 50D"
            ),
        ),
        migrations.AddField(
            model_name="option",
            name="open_interest_5d",
            field=models.FloatField(
                blank=True, help_text="Option Open Interest (5D)", null=True, verbose_name="Open Interest 5D"
            ),
        ),
        migrations.AddField(
            model_name="option",
            name="volatility",
            field=models.FloatField(blank=True, null=True, verbose_name="Volatility"),
        ),
        migrations.AddField(
            model_name="optionaggregate",
            name="open_interest_20d",
            field=models.FloatField(
                blank=True, help_text="Option Open Interest (20D)", null=True, verbose_name="Open Interest 20D"
            ),
        ),
        migrations.AddField(
            model_name="optionaggregate",
            name="open_interest_50d",
            field=models.FloatField(
                blank=True, help_text="Option Open Interest (50D)", null=True, verbose_name="Open Interest 50D"
            ),
        ),
        migrations.AddField(
            model_name="optionaggregate",
            name="open_interest_5d",
            field=models.FloatField(
                blank=True, help_text="Option Open Interest (5D)", null=True, verbose_name="Open Interest 5D"
            ),
        ),
        migrations.AddField(
            model_name="optionaggregate",
            name="volatility",
            field=models.FloatField(blank=True, null=True, verbose_name="Volatility"),
        ),
        migrations.AlterField(
            model_name="instrument",
            name="sedol",
            field=models.CharField(
                blank=True,
                help_text="Stock Exchange Daily Official List",
                max_length=255,
                null=True,
                verbose_name="SEDOL",
            ),
        ),
        migrations.AlterField(
            model_name="instrument",
            name="valoren",
            field=models.CharField(
                blank=True, help_text="Valoren Number", max_length=255, null=True, verbose_name="Valoren Number"
            ),
        ),
    ]



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class ClassificationEndpointConfig(EndpointViewConfig):
    pass


class ClassificationClassificationGroupEndpointConfig(ClassificationEndpointConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:classificationgroup-classification-list",
            args=[self.view.kwargs["group_id"]],
            request=self.request,
        )


class ClassificationParentClassificationEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:classificationparent-classification-list",
            args=[self.view.kwargs["parent_id"]],
            request=self.request,
        )


class ClassificationTreeChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:classificationgroup-treechart-list", args=[self.view.kwargs["group_id"]], request=self.request
        )


class ClassificationIcicleChartEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:classificationgroup-iciclechart-list",
            args=[self.view.kwargs["group_id"]],
            request=self.request,
        )


class InstrumentClassificationThroughEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrumentclassificationrelationship-list",
            args=[],
            request=self.request,
        )

    def get_update_endpoint(self, **kwargs):
        if "pk" in self.view.kwargs:
            return super().get_update_endpoint()
        return None


class InstrumentClassificationThroughInstrumentModelEndpointConfig(InstrumentClassificationThroughEndpointConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-classification-list",
            args=[self.view.instrument.get_root().id],
            request=self.request,
        )


class ClassificationInstrumentThroughInstrumentModelEndpointConfig(InstrumentClassificationThroughEndpointConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:classification-instrument-list",
            args=[self.view.kwargs["classification_id"]],
            request=self.request,
        )


class ClassificationInstrumentRelatedInstrumentEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:related_instrument-list",
            args=[self.view.kwargs["classified_instrument_id"]],
            request=self.request,
        )



from contextlib import suppress
from datetime import datetime
from typing import Any, Dict, Optional

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from wbcore.contrib.io.exceptions import DeserializationError
from wbcore.contrib.io.imports import ImportExportHandler

from .instrument import InstrumentImportHandler


class OptionAggregateImportHandler(ImportExportHandler):
    MODEL_APP_LABEL: str = "wbfdm.OptionAggregate"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instrument_handler = InstrumentImportHandler(self.import_source)

    def _deserialize(self, data):
        data["date"] = datetime.strptime(data["date"], "%Y-%m-%d").date()
        instrument = self.instrument_handler.process_object(data["instrument"], read_only=True)[0]
        if not instrument:
            raise DeserializationError("Instrument couldn't be found with given data")
        data["instrument"] = instrument

    def _get_instance(self, data: Dict[str, Any], history: Optional[models.QuerySet] = None, **kwargs) -> models.Model:
        with suppress(ObjectDoesNotExist):
            return self.model.objects.filter(instrument=data["instrument"], date=data["date"], type=data["type"])


class OptionImportHandler(ImportExportHandler):
    MODEL_APP_LABEL: str = "wbfdm.Option"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instrument_handler = InstrumentImportHandler(self.import_source)

    def _deserialize(self, data):
        data["date"] = datetime.strptime(data["date"], "%Y-%m-%d").date()
        data["expiration_date"] = datetime.strptime(data["expiration_date"], "%Y-%m-%d").date()
        instrument = self.instrument_handler.process_object(data["instrument"], read_only=True)[0]
        if not instrument:
            raise DeserializationError("Instrument couldn't be found with given data")
        data["instrument"] = instrument

    def _get_instance(self, data: Dict[str, Any], history: Optional[models.QuerySet] = None, **kwargs) -> models.Model:
        with suppress(ObjectDoesNotExist):
            return self.model.objects.filter(
                instrument=data["instrument"],
                contract_identifier=data["contract_identifier"],
                date=data["date"],
                type=data["type"],
            )



from datetime import timedelta
from decimal import Decimal

import pandas as pd
import pytest
from django.contrib.contenttypes.models import ContentType
from faker import Faker
from pandas.tseries.offsets import BDay

from wbfdm.models import Instrument, RelatedInstrumentThroughModel

fake = Faker()


@pytest.mark.django_db
class TestInstrumentPerformanceMetricBackend:
    def test_compute_metrics(self, weekday, instrument, instrument_price_factory):
        from wbfdm.contrib.metric.backends.performances import (  # we import locally to avoid database access pytest error
            InstrumentPerformanceMetricBackend,
        )

        backend = InstrumentPerformanceMetricBackend(weekday)

        price_today = instrument_price_factory.create(instrument=instrument, calculated=False, date=weekday)
        price_yesterday = instrument_price_factory.create(
            instrument=instrument, calculated=False, date=weekday - BDay(1)
        )
        price_last_week = instrument_price_factory.create(
            instrument=instrument, calculated=False, date=(weekday - timedelta(days=7)) - BDay(0)
        )
        price_last_month = instrument_price_factory.create(
            instrument=instrument, calculated=False, date=(weekday - timedelta(days=30)) - BDay(1)
        )
        price_last_quarter = instrument_price_factory.create(
            instrument=instrument, calculated=False, date=(weekday - timedelta(days=120)) - BDay(1)
        )
        price_last_year = instrument_price_factory.create(
            instrument=instrument, calculated=False, date=(weekday - timedelta(days=365)) - BDay(1)
        )
        price_end_last_year = instrument_price_factory.create(
            instrument=instrument, calculated=False, date=weekday - pd.tseries.offsets.BYearEnd(1)
        )
        price_inception = instrument_price_factory.create(
            instrument=instrument, calculated=False, date=price_last_year.date - BDay(1)
        )

        instrument_price_factory.create(
            instrument=instrument, calculated=True, date=price_inception.date - BDay(1)
        )  # add noise
        instrument.inception_date = price_inception.date.date()
        instrument.save()
        instrument.update_last_valuation_date()
        base_metric = next(backend.compute_metrics(instrument))

        # Check base metric metadata
        assert base_metric.date is None
        assert base_metric.basket_id == instrument.id
        assert base_metric.basket_content_type_id == ContentType.objects.get_for_model(Instrument).id
        assert base_metric.key == "performance"
        assert base_metric.instrument_id is None

        assert base_metric.metrics["is_estimated"] is False
        assert base_metric.metrics["daily"] == pytest.approx(
            float(price_today.net_value / price_yesterday.net_value - Decimal(1)), abs=1e-5
        )
        assert base_metric.metrics["weekly"] == pytest.approx(
            float(price_today.net_value / price_last_week.net_value - Decimal(1)), abs=1e-5
        )
        assert base_metric.metrics["monthly"] == pytest.approx(
            float(price_today.net_value / price_last_month.net_value - Decimal(1)), abs=1e-5
        )
        assert base_metric.metrics["quarterly"] == pytest.approx(
            float(price_today.net_value / price_last_quarter.net_value - Decimal(1)), abs=1e-5
        )

        assert base_metric.metrics["yearly"] == pytest.approx(
            float(price_today.net_value / price_last_year.net_value - Decimal(1)), abs=1e-5
        )

        assert base_metric.metrics["year_to_date"] == pytest.approx(
            float(price_today.net_value / price_end_last_year.net_value - Decimal(1)), abs=1e-5
        )
        assert base_metric.metrics["inception"] == pytest.approx(
            float(price_today.net_value / price_inception.net_value - Decimal(1)), abs=1e-5
        )

    def test_compute_metrics_estimated(self, weekday, instrument, instrument_price_factory):
        from wbfdm.contrib.metric.backends.performances import (  # we import locally to avoid database access pytest error
            InstrumentPerformanceMetricBackend,
        )

        backend = InstrumentPerformanceMetricBackend(weekday)

        price_today = instrument_price_factory.create(instrument=instrument, calculated=True, date=weekday)
        price_yesterday = instrument_price_factory.create(
            instrument=instrument, calculated=False, date=weekday - BDay(1)
        )
        instrument.inception_date = price_yesterday.date.date()
        instrument.save()
        instrument.update_last_valuation_date()
        base_metric = next(backend.compute_metrics(instrument))

        assert base_metric.metrics["is_estimated"] is True
        assert base_metric.metrics["daily"] == pytest.approx(
            float(price_today.net_value / price_yesterday.net_value - Decimal(1)), abs=1e-5
        )

    def test_compute_metrics_with_peers(self, weekday, instrument_factory, instrument_price_factory):
        from wbfdm.contrib.metric.backends.performances import (  # we import locally to avoid database access pytest error
            InstrumentPerformanceMetricBackend,
        )

        instrument = instrument_factory.create()
        peer_1 = instrument_factory.create()
        RelatedInstrumentThroughModel.objects.create(
            related_type="PEER", instrument=instrument, related_instrument=peer_1
        )
        peer_2 = instrument_factory.create()
        RelatedInstrumentThroughModel.objects.create(
            related_type="PEER", instrument=instrument, related_instrument=peer_2
        )

        backend = InstrumentPerformanceMetricBackend(weekday)

        price_today = instrument_price_factory.create(instrument=instrument, calculated=False, date=weekday)
        price_yesterday = instrument_price_factory.create(
            instrument=instrument, calculated=False, date=weekday - BDay(1)
        )
        instrument.inception_date = price_yesterday.date.date()
        instrument.save()
        peer_1_price_today = instrument_price_factory.create(instrument=peer_1, calculated=False, date=weekday)
        peer_1_price_yesterday = instrument_price_factory.create(
            instrument=peer_1, calculated=False, date=weekday - BDay(1)
        )
        peer_1.inception_date = peer_1_price_yesterday.date.date()
        peer_1.save()
        peer_2_price_today = instrument_price_factory.create(instrument=peer_2, calculated=False, date=weekday)
        peer_2_price_yesterday = instrument_price_factory.create(
            instrument=peer_2, calculated=False, date=weekday - BDay(1)
        )
        peer_2.inception_date = peer_2_price_yesterday.date.date()
        peer_2.save()
        instrument.update_last_valuation_date()
        peer_1.update_last_valuation_date()
        peer_2.update_last_valuation_date()
        base_metric = next(backend.compute_metrics(instrument))
        daily_instrument_perf = float(price_today.net_value / price_yesterday.net_value - Decimal(1))
        assert base_metric.metrics["daily"] == pytest.approx(daily_instrument_perf, abs=1e-5)
        daily_peer_1_perf = float(peer_1_price_today.net_value / peer_1_price_yesterday.net_value - Decimal(1))
        daily_peer_2_perf = float(peer_2_price_today.net_value / peer_2_price_yesterday.net_value - Decimal(1))
        avg_daily_peer_perf = (daily_peer_1_perf + daily_peer_2_perf) / 2
        r = daily_instrument_perf - avg_daily_peer_perf
        assert base_metric.metrics["peer_daily"] == pytest.approx(r, abs=0.00001)



from django.contrib import admin

from .instruments import InstrumentModelAdmin, InstrumentTypeModelAdmin
from .classifications import ClassificationAdmin, ClassificationGroupAdmin
from .exchanges import ExchangeModelAdmin
from .instrument_lists import InstrumentListAdmin
from .instruments_relationships import InstrumentFavoriteGroupModelAdmin
from .esg import ControversyModelAdmin
from .exchanges import ExchangeModelAdmin
from .instrument_prices import InstrumentPriceAdmin
from .instrument_requests import InstrumentRequestModelAdmin
from .options import OptionModelAdmin, OptionAggregateModelAdmin


class ReadOnlyAdmin(admin.ModelAdmin):
    readonly_fields = []

    def get_readonly_fields(self, request, obj=None):
        return (
            list(self.readonly_fields)
            + [field.name for field in obj._meta.fields]  # type: ignore
            + [field.name for field in obj._meta.many_to_many]  # type: ignore
        )

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class ReadOnlyTabularInline(admin.TabularInline):
    show_change_link = True

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False



from django.contrib.contenttypes.models import ContentType
from django.utils.translation import gettext as _
from wbcore import filters

from wbfdm.contrib.metric.models import InstrumentMetric

from .registry import backend_registry


class InstrumentMetricFilterSet(filters.FilterSet):
    key = filters.ChoiceFilter(choices=backend_registry.get_choices(), label="Key")
    basket_content_type = filters.ModelChoiceFilter(
        queryset=ContentType.objects.all(),
        endpoint="wbcore:contenttyperepresentation-list",
        value_key="id",
        label_key="{{app_label}} | {{model}}",
        label=_("Basket Content Type"),
    )

    class Meta:
        model = InstrumentMetric
        fields = {
            "basket_content_type": ["exact"],
            "basket_id": ["exact"],
            "instrument": ["exact"],
            "key": ["exact"],
            "date": ["lte", "gte", "exact"],
            "parent_metric": ["exact", "isnull"],
        }



from django.db.models import Case, Exists, F, IntegerField, OuterRef, When
from wbcore import viewsets

from wbfdm.contrib.metric.filters import InstrumentMetricFilterSet
from wbfdm.contrib.metric.models import InstrumentMetric
from wbfdm.contrib.metric.serializers import (
    InstrumentMetricModelSerializer,
    InstrumentMetricRepresentationSerializer,
)
from wbfdm.contrib.metric.viewsets.configs import InstrumentMetricDisplayConfig


class InstrumentMetricRepresentationViewSet(viewsets.RepresentationViewSet):
    ordering = ["-date"]
    ordering_fields = [
        "key",
        "date",
    ]
    filterset_class = InstrumentMetricFilterSet

    serializer_class = InstrumentMetricRepresentationSerializer
    queryset = InstrumentMetric.objects.all()


class InstrumentMetricViewSet(viewsets.ModelViewSet):
    queryset = InstrumentMetric.objects.select_related("basket_content_type", "instrument").annotate(
        has_children=Exists(InstrumentMetric.objects.filter(parent_metric=OuterRef("pk"))),
        _group_key=Case(When(has_children=True, then=F("id")), default=None, output_field=IntegerField()),
    )

    serializer_class = InstrumentMetricModelSerializer
    filterset_class = InstrumentMetricFilterSet

    display_config_class = InstrumentMetricDisplayConfig

    ordering = ["-date"]
    ordering_fields = [
        "key",
        "date",
    ]
    search_fields = ["key", "instrument__computed_str", "basket_repr"]



import numpy as np
import pandas as pd

from wbfdm.import_export.backends.refinitiv.instrument_price import DEFAULT_MAPPING

from .utils import _clean_and_return_dict


def parse(import_source):
    # Read Json as dataframe
    df = pd.read_json(import_source.file, orient="records")
    df = df.replace([np.inf, -np.inf, np.nan], None)
    # Sanitize df and rename columns to model knowledge's
    df = (
        df.replace({"NA": None})
        .rename(columns=DEFAULT_MAPPING)
        .rename(columns={"Instrument": "instrument", "Dates": "date"})
    )
    # Sanitize
    df = df.replace([np.inf, -np.inf, np.nan], None)
    df = df.dropna(how="all", subset=df.columns.difference(["instrument"]))

    df["date"] = pd.to_datetime(df["date"], utc=True, unit="ms")
    df = df.sort_values(by="date", ascending=True)
    df["date"] = df["date"].dt.strftime("%Y-%m-%d")
    df = df.replace([np.inf, -np.inf, np.nan], None)
    return {"data": _clean_and_return_dict(df)}



# from wbcore.test import default_config
#
# config = {}
# for key, value in default_config.items():
#     config[key] = list(
#         filter(
#             lambda x: x.__module__.startswith("wbfdm") and False,
#             value
#         )
#     )



from typing import Optional

from wbcore.contrib.color.enums import WBColor
from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import DisplayViewConfig

perf_formatting_rules = [
    dp.FormattingRule(
        style={
            "color": WBColor.RED_DARK.value,
        },
        condition=("<", 0),
    ),
    dp.FormattingRule(
        style={
            "color": WBColor.GREEN_DARK.value,
        },
        condition=(">", 0),
    ),
]
perf_formatting_rules_bold = [
    dp.FormattingRule(
        style={
            "color": WBColor.RED_DARK.value,
            "fontWeight": "bold",
        },
        condition=("<", 0),
    ),
    dp.FormattingRule(
        style={
            "color": WBColor.GREEN_DARK.value,
            "fontWeight": "bold",
        },
        condition=(">", 0),
    ),
]


class MonthlyPerformancesInstrumentDisplayViewConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(
                    key="year",
                    label="Year",
                    width=Unit.PIXEL(75),
                    formatting_rules=[
                        dp.FormattingRule(
                            style={
                                "fontWeight": "bold",
                            },
                        ),
                    ],
                ),
                dp.Field(key="1", label="Jan", width=Unit.PIXEL(75), formatting_rules=perf_formatting_rules),
                dp.Field(key="2", label="Feb", width=Unit.PIXEL(75), formatting_rules=perf_formatting_rules),
                dp.Field(key="3", label="Mar", width=Unit.PIXEL(75), formatting_rules=perf_formatting_rules),
                dp.Field(key="4", label="Apr", width=Unit.PIXEL(75), formatting_rules=perf_formatting_rules),
                dp.Field(key="5", label="May", width=Unit.PIXEL(75), formatting_rules=perf_formatting_rules),
                dp.Field(key="6", label="Jun", width=Unit.PIXEL(75), formatting_rules=perf_formatting_rules),
                dp.Field(key="7", label="Jul", width=Unit.PIXEL(75), formatting_rules=perf_formatting_rules),
                dp.Field(key="8", label="Aug", width=Unit.PIXEL(75), formatting_rules=perf_formatting_rules),
                dp.Field(key="9", label="Sep", width=Unit.PIXEL(75), formatting_rules=perf_formatting_rules),
                dp.Field(key="10", label="Oct", width=Unit.PIXEL(75), formatting_rules=perf_formatting_rules),
                dp.Field(key="11", label="Nov", width=Unit.PIXEL(75), formatting_rules=perf_formatting_rules),
                dp.Field(key="12", label="Dec", width=Unit.PIXEL(75), formatting_rules=perf_formatting_rules),
                dp.Field(
                    key="annual", label="Yearly", width=Unit.PIXEL(75), formatting_rules=perf_formatting_rules_bold
                ),
            ]
        )



from typing import Iterator

from django.db import connections
from jinjasql import JinjaSql
from wbcore.contrib.dataloader.dataloaders import Dataloader
from wbcore.contrib.dataloader.utils import dictfetchall

from wbfdm.dataloaders.protocols import ReportDateProtocol
from wbfdm.dataloaders.types import ReportDateDataDict


class IbesReportingDateDataloader(ReportDateProtocol, Dataloader):
    def reporting_dates(self, only_next: bool = True) -> Iterator[ReportDateDataDict]:
        lookup = {k: v for k, v in self.entities.values_list("dl_parameters__reporting_dates__parameters", "id")}

        sql = """
            with next_events as (
                select
                    *,
                    row_number() over (partition by EstPermID, PerType order by PerEndDate) as rn
                from TREExpectedRptDate
                where StartDate > getdate()
            )

            select
                rp.EstPermID as external_id,
                'qa-ibes' as source,
                convert(date, rp.PerEndDate) as per_end_date,
                convert(date, rp.StartDate) as start_date,
                convert(date, rp.EndDate) as end_date,
                iif(rp.PerType=4, 0, 1) as interim,
                case
                    when rp.MarketPhase = 'AMC' then 'after_market'
                    when rp.MarketPhase = 'BMO' then 'before_market'
                    else null
                end as market_phase,
                lower(rp.Status) as status
            {% if only_next %}
                from next_events as rp
            {% else %}
                from TREExpectedRptDate as rp
            {% endif %}
            where rp.EstPermID in (
            {% for instrument in instruments %}
                {{instrument}} {% if not loop.last %}, {% endif %}
            {% endfor %}
            )
            {% if only_next %}
                AND rp.rn = 1
            {% endif %}
        """

        query, bind_params = JinjaSql(param_style="format").prepare_query(
            sql,
            {
                "instruments": lookup.keys(),
                "only_next": only_next,
            },
        )
        with connections["qa"].cursor() as cursor:
            cursor.execute(
                query,
                bind_params,
            )
            for row in dictfetchall(cursor):
                row["instrument_id"] = lookup[row["external_id"]]
                row["interim"] = bool(row["interim"])
                yield row



from typing import List, Optional

import roman
from django.contrib.postgres.expressions import ArraySubquery
from django.db import models
from django.db.models.functions import Cast
from django.db.models.signals import post_save
from django.dispatch import receiver
from mptt.models import MPTTModel, TreeForeignKey
from wbcore.models import WBModel
from wbcore.utils.models import ComplexToStringMixin

from .instrument_relationships import InstrumentClassificationThroughModel


class ClassificationGroup(WBModel):
    name = models.CharField(max_length=128, verbose_name="Name")
    is_primary = models.BooleanField(
        default=False,
        verbose_name="Primary",
        help_text="Set to True if this " "classification must be used as " "default if not specified " "otherwise",
    )
    max_depth = models.IntegerField(default=0, verbose_name="Maximum Depth")
    code_level_digits = models.IntegerField(default=2, verbose_name="The number of digits per code level")

    def __str__(self) -> str:
        return f'{self.name} ({"Primary" if self.is_primary else "Non Primary"})'

    def save(self, *args, **kwargs):
        qs = ClassificationGroup.objects.filter(is_primary=True).exclude(id=self.id)
        if self.is_primary:
            qs.update(is_primary=False)
        elif not qs.exists():
            self.is_primary = True
        return super().save(*args, **kwargs)

    def get_levels_representation(self) -> List[str]:
        return [
            self.classifications.filter(height=i).first().level_representation
            for i in range(self.max_depth + 1)
            if self.classifications.filter(height=i).exists()
        ]

    def get_fields_names(self, sep: str = "__") -> list[str]:
        return [f"parent{f'{sep}parent' * height}" for height in range(self.max_depth)]

    def annotate_queryset(
        self,
        queryset: models.QuerySet,
        classification_height: int,
        instrument_label_key: str,
        unique: bool = False,
        annotation_label: str = "classifications",
    ):
        ref_id = "classification__"
        if classification_height:
            ref_id += f"{'parent__' * classification_height}"
        ref_id += "id"
        if instrument_label_key:
            instrument_label_key += "__"
        base_subquery = InstrumentClassificationThroughModel.objects.filter(
            classification__group=self,
            instrument__tree_id=models.OuterRef(f"{instrument_label_key}tree_id"),
            instrument__lft__lte=models.OuterRef(f"{instrument_label_key}lft"),
            instrument__rght__gte=models.OuterRef(f"{instrument_label_key}rght"),
        )
        if unique:
            expression = models.Subquery(base_subquery.values(ref_id)[:1])
        else:
            expression = ArraySubquery(
                base_subquery.values(ref_id).distinct(ref_id)
            )  # we use distinct in order to remove duplicated classification (e.g. classification added to the parent as well as the children)

        return queryset.annotate(**{annotation_label: expression})

    class Meta:
        verbose_name = "Classification Group"
        verbose_name_plural = "Classification Groups"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbfdm:classificationgrouprepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbfdm:classificationgroup"


class Classification(MPTTModel, ComplexToStringMixin):
    parent = TreeForeignKey(
        "self",
        related_name="children",
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        verbose_name="Parent Classification",
    )
    height = models.PositiveIntegerField(
        default=0,
        verbose_name="The height (leaf node have height 0)",
    )
    group = models.ForeignKey(
        ClassificationGroup,
        related_name="classifications",
        on_delete=models.CASCADE,
        verbose_name="Classification Group",
    )

    level_representation = models.CharField(max_length=256, verbose_name="Level Representation")

    name = models.CharField(max_length=128, verbose_name="Name")
    code_aggregated = models.CharField(max_length=64, verbose_name="Code Aggregated")

    investable = models.BooleanField(default=True, help_text="Is this classification investable for us?")

    description = models.TextField(
        default="",
        blank=True,
        help_text="Give a basic definition and description",
        verbose_name="Definition/Description",
    )

    @classmethod
    def dict_to_model(cls, classification_data):
        if isinstance(classification_data, int):
            return cls.objects.filter(id=classification_data).first()
        res = cls.objects.all()
        if code_aggregated := classification_data.get("code_aggregated", None):
            res = res.filter(code_aggregated=code_aggregated)
        if group_id := classification_data.get("group", None):
            res = res.filter(group=group_id)
        if res.count() == 1:
            return res.first()

    def __str__(self):
        if self.computed_str:
            return self.computed_str
        return f"{self.code_aggregated} {self.name}"

    def get_classified_instruments(self, only_favorites: bool = False) -> models.QuerySet:
        childs_classifications = self.get_descendants(include_self=True)
        params = {"classifications__in": childs_classifications}
        if only_favorites:
            params["classifications_through__is_favorite"] = True

        from wbfdm.models import Instrument

        return Instrument.objects.filter(**params).distinct()

    def save(self, *args, **kwargs):
        if self.parent:
            self.group = self.parent.group
        if not self.code_aggregated:
            self.code_aggregated = self.get_next_valid_code(self.group, self.parent)
        if not self.level_representation:
            self.level_representation = self.get_default_level_representation(self.group, self.parent)
        return super().save(*args, **kwargs)

    def compute_str(self) -> str:
        if parent := self.parent:
            tree_titles = parent.name
            while parent and (parent := parent.parent):
                tree_titles += f" - {parent.name}"
            return f"{self.code_aggregated} {self.name} [{tree_titles}] ({self.group.name})"
        return f"{self.code_aggregated} {self.name} ({self.group.name})"

    class Meta:
        verbose_name = "Classification"
        verbose_name_plural = "Classifications"
        constraints = [models.UniqueConstraint(fields=["group", "code_aggregated"], name="unique_classification")]

    @classmethod
    def get_next_valid_code(cls, group: "ClassificationGroup", parent: "Classification | None" = None) -> str:
        """
        Return the next valid classification code given the classification parent and its group parameters
        Args:
            group: The classification group the estimated code belongs to
            parent: The classification parent (if any)

        Returns:
            The next valid and unused aggregated classification code
        """
        if not group:
            raise ValueError("This method needs a group")
        siblings_classifications = (
            Classification.objects.filter(parent=parent, group=group)
            .annotate(casted_code=Cast("code_aggregated", models.BigIntegerField()))
            .order_by("-casted_code")
        )
        parent_level = parent.level + 1 if parent else 0
        code_aggregated_digits = group.code_level_digits * (parent_level + 1)
        if last_classification := siblings_classifications.first():
            for i in range(0, 100 - last_classification.casted_code % 10**group.code_level_digits):
                if not siblings_classifications.filter(casted_code=last_classification.casted_code + i).exists():
                    last_valid_code = last_classification.casted_code + i
                    return f"{last_valid_code:0{code_aggregated_digits}}"
        if parent:
            return parent.code_aggregated + f"{1:0{group.code_level_digits}}"
        return f"{1:0{group.code_level_digits}}"

    @classmethod
    def get_default_level_representation(
        cls, group: "ClassificationGroup", parent: Optional["Classification"] = None
    ) -> str:
        """
        Return the default level representation, extracted from the classification siblings
        Args:
            group: The classification group the estimated code belongs to
            parent: The classification parent (if any)

        Returns:
            A default level representation (e.g. Industry)
        """
        level = parent.level + 1 if parent else None
        siblings_classifications = cls.objects.filter(level=level, group=group).order_by("id")
        if siblings_classifications.exists():
            return siblings_classifications.last().level_representation
        level = roman.toRoman(level) if level else 0
        return f"Level {level}"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbfdm:classificationrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbfdm:classification"


@receiver(post_save, sender="wbfdm.Classification")
def post_save_parent_classification(sender, instance, created, raw, **kwargs):
    # Recursively call the parent save function to recompute its parameters
    if not raw and instance.parent:
        instance.parent.save()
    if instance.level is not None:
        update_fields = {"height": instance.group.max_depth - instance.level}
        Classification.objects.filter(id=instance.id).update(**update_fields)
        instance.refresh_from_db()
    # # Ensure initial parent classifcation span the proper classification tree structure
    if instance.group and not instance.get_descendants().exists() and instance.level < instance.group.max_depth:
        Classification.objects.create(
            parent=instance,
            group=instance.group,
            name=f"{instance.name} (Level {instance.level})",
        )
    """
    If a parent classification is not investable, therefore its "children" become non investable too, by cascade.
    If a child classification becomes investable and one of its parent is non investable, therefore he cannot be
    investable as long as its parent is non investable. (-> in the serializer)
    """
    if not raw and not instance.investable and instance.get_descendants().exists():
        descandants = instance.get_descendants()
        descandants.update(investable=False)



import factory
from django.contrib.contenttypes.models import ContentType
from faker import Faker

from wbfdm.contrib.metric.models import InstrumentMetric
from wbfdm.factories import InstrumentFactory
from wbfdm.models import Instrument

from .registry import backend_registry

fake = Faker()


def _get_metrics(key: str):
    backend = backend_registry[key, Instrument]
    metrics = {}
    for sub_field in backend.keys[0].subfields:
        metrics[sub_field.key] = fake.pyfloat()
    return metrics


class InstrumentMetricFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = InstrumentMetric
        # django_get_or_create = ["basket_content_type", "basket_id", "instrument", "date", "key"]

    basket = factory.SubFactory(InstrumentFactory)
    basket_content_type = factory.LazyAttribute(lambda o: ContentType.objects.get_for_model(Instrument))
    basket_id = factory.LazyAttribute(lambda o: o.basket.id)

    date = factory.Faker("date_object")
    key = factory.Iterator(backend_registry.keys())
    metrics = factory.LazyAttribute(lambda o: _get_metrics(o.key))
    parent_metric = None



from django import forms
from django.contrib import admin

from wbfdm.models import Exchange


class CsvImportForm(forms.Form):
    csv_file = forms.FileField()


@admin.register(Exchange)
class ExchangeModelAdmin(admin.ModelAdmin):
    list_display = (
        "source_id",
        "name",
        "mic_code",
        "operating_mic_code",
        "bbg_exchange_codes",
        "bbg_composite",
        "country",
    )
    search_fields = (
        "name",
        "operating_mic_code",
        "mic_code",
        "bbg_exchange_codes",
        "bbg_composite",
        "refinitiv_identifier_code",
        "comments",
    )
    list_filter = ["country"]
    fieldsets = (
        (
            "Main Data",
            {
                "fields": (
                    (
                        "name",
                        "opening_time",
                        "closing_time",
                    ),
                    ("mic_code", "mic_name"),
                    ("operating_mic_code", "operating_mic_name"),
                    ("bbg_exchange_codes", "bbg_composite_primary", "bbg_composite"),
                    ("refinitiv_identifier_code", "refinitiv_mnemonic"),
                    ("country", "website"),
                    ("comments",),
                )
            },
        ),
    )

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("country")



# Generated by Django 5.0.9 on 2024-12-19 12:56

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0027_remove_instrument_unique_ric_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="instrumentprice",
            name="annualized_daily_volatility",
            field=models.FloatField(blank=True, null=True, verbose_name="Annualized Volatility"),
        ),
    ]



from unittest.mock import patch

from faker import Faker

from ..dispatch import compute_metrics
from ..orchestrators import MetricOrchestrator
from ..tasks import compute_metrics_as_task

fake = Faker()


@patch.object(MetricOrchestrator, "process")
def test_dispatch_compute_metrics(mock_fct):
    key = "performance"
    basket = object()
    val_date = fake.date_object()
    compute_metrics(val_date, key=key, basket=basket)
    mock_fct.assert_called_once()


@patch.object(MetricOrchestrator, "process")
def test_periodically_compute_all_metrics(mock_fct):
    compute_metrics_as_task()
    mock_fct.assert_called_once()



from wbcore import filters as wb_filters

from wbfdm.filters.utils import _get_default_classification_group_id
from wbfdm.models import (
    Classification,
    ClassificationGroup,
    Instrument,
    InstrumentClassificationThroughModel,
)


class ClassificationFilter(wb_filters.FilterSet):
    instruments = wb_filters.ModelChoiceFilter(
        label="Instrument",
        queryset=Instrument.objects.all(),
        endpoint=Instrument.get_representation_endpoint(),
        filter_params={"is_classifiable": True},
        value_key=Instrument.get_representation_value_key(),
        label_key=Instrument.get_representation_label_key(),
    )

    instruments_neq = wb_filters.ModelChoiceFilter(
        label="Instrument not classified in",
        queryset=Instrument.objects.all(),
        endpoint=Instrument.get_representation_endpoint(),
        value_key=Instrument.get_representation_value_key(),
        label_key=Instrument.get_representation_label_key(),
        filter_params={"is_classifiable": True},
        field_name="instruments",
        lookup_expr="exact",
        exclude=True,
    )

    instrument_type_key = wb_filters.CharFilter(
        label="Instrument Type Key", hidden=True, method="filter_instrument_type_key"
    )

    def filter_instrument_type_key(self, queryset, name, value):
        if value:
            return queryset.filter(instruments__instrument_type__key=value).distinct()
        return queryset

    class Meta:
        model = Classification
        fields = {
            "id": ["in"],
            "parent": ["exact"],
            "height": ["gte", "exact", "lte"],
            "level": ["gte", "exact", "lte"],
            "group": ["exact"],
            "level_representation": ["icontains"],
            "name": ["icontains"],
            "code_aggregated": ["icontains", "exact"],
        }
        hidden_fields = ["id__in"]


class ClassificationTreeChartFilter(wb_filters.FilterSet):
    top_classification = wb_filters.ModelChoiceFilter(
        label="Top Classification",
        queryset=Classification.objects.all(),
        endpoint=Classification.get_representation_endpoint(),
        value_key=Classification.get_representation_value_key(),
        label_key=Classification.get_representation_label_key(),
        method="filter_top_classification",
    )

    aggregation_type = wb_filters.ChoiceFilter(
        choices=[("classification_count", "Classification Count"), ("instrument_count", "Instrument Count")],
        default="classification_count",
        label="Aggregation Type",
        method="fake_filter",
    )

    def filter_top_classification(self, queryset, name, value):
        if value:
            return queryset.filter(id__in=value.get_descendants().values("id"))
        return queryset

    class Meta:
        model = Classification
        fields = {}


class InstrumentClassificationThroughModelViewFilterSet(wb_filters.FilterSet):
    classification__group = wb_filters.ModelChoiceFilter(
        label="Group",
        queryset=ClassificationGroup.objects.all(),
        endpoint=ClassificationGroup.get_representation_endpoint(),
        value_key=ClassificationGroup.get_representation_value_key(),
        label_key=ClassificationGroup.get_representation_label_key(),
        default=_get_default_classification_group_id,
    )

    class Meta:
        model = InstrumentClassificationThroughModel
        fields = {
            "instrument": ["exact"],
            "classification": ["exact"],
            "is_favorite": ["exact"],
            "pure_player": ["exact"],
            "top_player": ["exact"],
        }



import pytz
from django.conf import settings
from django.db import connections
from django.db.models import Max
from wbcore.contrib.dataloader.utils import dictfetchall, dictfetchone
from wbcore.contrib.geography.models import Geography
from wbcore.utils.cache import mapping
from wbfdm.models import Exchange
from wbfdm.sync.abstract import Sync


class QAExchangeSync(Sync[Exchange]):
    SOURCE = "qa-ds2-exchange"

    def update(self):
        sql = "SELECT ExchIntCode AS source_id, ExchName AS name, ExchCtryCode AS country_id, ExchMnem AS refinitiv_mnemonic FROM DS2Exchange"
        Exchange.objects.bulk_create(
            map(
                lambda data: Exchange(
                    source=self.SOURCE,
                    source_id=data["source_id"],
                    name=data["name"],
                    country_id=mapping(Geography.countries, "code_2").get(data["country_id"]),
                    refinitiv_mnemonic=data["refinitiv_mnemonic"],
                ),
                dictfetchall(connections["qa"].cursor().execute(sql)),
            ),
            update_conflicts=True,
            update_fields=["name", "country", "refinitiv_mnemonic"],
            unique_fields=["source", "source_id"],
        )

    def update_or_create_item(self, external_id: int) -> Exchange:
        defaults = self.get_item(external_id)
        defaults["country_id"] = mapping(Geography.countries, "code_2").get(defaults["country_id"])
        exchange, _ = Exchange.objects.update_or_create(
            source=self.SOURCE,
            source_id=external_id,
            defaults=defaults,
        )
        return exchange

    def update_item(self, item: Exchange) -> Exchange:
        return self.update_or_create_item(item.source_id)

    def get_item(self, external_id: int) -> dict:
        sql = "SELECT ExchName AS name, ExchCtryCode AS country_id, ExchMnem AS refinitiv_mnemonic FROM DS2Exchange WHERE ExchIntCode = %s"
        return dictfetchone(connections["qa"].cursor().execute(sql, (external_id,)))

    def trigger_partial_update(self):
        max_last_updated = (
            Exchange.objects.filter(source=self.SOURCE)
            .aggregate(max_last_updated=Max("last_updated"))
            .get("max_last_updated")
        )
        if max_last_updated is None:
            self.update()
        else:
            with connections["qa"].cursor() as cursor:
                cursor.execute(
                    "SELECT MAX(last_user_update) FROM sys.dm_db_index_usage_stats WHERE OBJECT_NAME(object_id) = 'DS2Exchange_changes'"
                )
                max_last_updated_qa = (
                    pytz.timezone(settings.TIME_ZONE).localize(result[0]) if (result := cursor.fetchone()) else None
                )
                if max_last_updated_qa and max_last_updated_qa > max_last_updated:
                    for _, exchange_id in cursor.execute(
                        "SELECT UpdateFlag_, ExchIntCode FROM DS2Exchange_changes"
                    ).fetchall():
                        self.update_or_create_item(exchange_id)



from wbcore.metadata.configs.titles import TitleViewConfig


class InstrumentFavoriteGroupTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Instrument Favorite Groups"

    def get_instance_title(self):
        try:
            obj = self.view.get_object()
            return f"Favorite Group: {str(obj)}"
        except AssertionError:
            return "Instrument: {{name}}"

    def get_create_title(self):
        return "New Favorite Group"


class ClassifiedInstrumentTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        return "Classified Instruments"



from .classifications import (
    ClassificationClassificationGroupEndpointConfig,
    ClassificationEndpointConfig,
    ClassificationIcicleChartEndpointConfig,
    ClassificationInstrumentRelatedInstrumentEndpointConfig,
    ClassificationInstrumentThroughInstrumentModelEndpointConfig,
    ClassificationParentClassificationEndpointConfig,
    ClassificationTreeChartEndpointConfig,
    InstrumentClassificationThroughEndpointConfig,
    InstrumentClassificationThroughInstrumentModelEndpointConfig,
)
from .financials_analysis import (
    CashFlowAnalysisInstrumentBarChartEndpointConfig,
    CashFlowAnalysisInstrumentTableChartEndpointConfig,
    EarningsInstrumentChartEndpointConfig,
    FinancialsGraphInstrumentChartEndpointConfig,
    NetDebtAndEbitdaInstrumentChartEndpointConfig,
    ProfitabilityRatiosInstrumentChartEndpointConfig,
    SummaryTableInstrumentChartEndpointConfig,
)
from .instrument_lists import (
    InstrumentListThroughModelInstrumentListEndpointConfig,
    InstrumentListThroughModelInstrumentEndpointConfig,
    InstrumentListThroughModelEndpointConfig,
)
from .instrument_requests import InstrumentRequestEndpointConfig
from .instruments import InstrumentEndpointConfig
from .instrument_prices import (
    InstrumentPriceInstrumentEndpointConfig,
    InstrumentPriceStatisticsInstrumentEndpointConfig,
    MonthlyPerformancesInstrumentEndpointConfig,
    FinancialStatisticsInstrumentEndpointConfig,
    InstrumentPriceInstrumentDistributionReturnsChartEndpointConfig,
    BestAndWorstReturnsInstrumentEndpointConfig,
)
from .instruments_relationships import (
    ClassifiedInstrumentEndpointConfig,
    RelatedInstrumentThroughInstrumentEndpointConfig,
    InstrumentFavoriteGroupEndpointConfig
)
from .exchanges import ExchangeEndpointConfig
from .esg import InstrumentESGPAIEndpointViewConfig, InstrumentESGControversiesEndpointViewConfig



from django.db.models import (
    AutoField,
    Exists,
    ExpressionWrapper,
    F,
    OuterRef,
    QuerySet,
    Subquery,
)
from django.db.models.functions import Coalesce


class InstrumentQuerySet(QuerySet):
    def annotate_classification_for_group(
        self, classification_group, classification_height: int = 0, **kwargs
    ) -> QuerySet:
        return classification_group.annotate_queryset(
            self, classification_height, "", annotation_label="ancestor_classifications", **kwargs
        )

    def annotate_base_data(self):
        base_qs = InstrumentQuerySet(self.model, using=self._db)
        return self.annotate(
            is_investable=~Exists(base_qs.filter(parent=OuterRef("pk"))),
            root=Subquery(base_qs.filter(tree_id=OuterRef("tree_id"), level=0).values("id")[:1]),
            primary_security=ExpressionWrapper(
                Coalesce(
                    Subquery(
                        base_qs.filter(
                            parent=OuterRef("pk"),
                            is_primary=True,
                            is_security=True,
                        ).values(
                            "id"
                        )[:1]
                    ),
                    F("id"),
                ),
                output_field=AutoField(),
            ),
            primary_quote=ExpressionWrapper(
                Coalesce(
                    Subquery(
                        base_qs.filter(
                            parent=OuterRef("primary_security"),
                            is_primary=True,
                        ).values(
                            "id"
                        )[:1]
                    ),
                    F("primary_security"),
                ),
                output_field=AutoField(),
            ),
        )

    def annotate_all(self):
        return self.annotate_base_data()

    @property
    def dl(self):
        """Provides access to the dataloader proxy for the entities in the QuerySet.

        This method allows for easy retrieval of the DataloaderProxy instance
        associated with the QuerySet. It enables the utilization of dataloader
        functionalities directly from the QuerySet, facilitating data fetching and
        processing tasks.

        Returns:
            DataloaderProxy: An instance of DataloaderProxy associated with the
                entities in the QuerySet.
        """
        return self.model.dl_proxy(self)



from wbcore.metadata.configs.titles import TitleViewConfig


class StatementTitleViewConfig(TitleViewConfig):
    def get_list_title(self):
        try:
            statement_label = self.view.financial_analysis_mapping[self.view.kwargs["statement"]][1]
            return f"{self.view.instrument} - {statement_label} Table"
        except KeyError:
            return f"{self.view.instrument} - Statement"



import logging
from typing import Any

from celery import shared_task
from django.contrib.contenttypes.models import ContentType
from langchain_core.messages import HumanMessage, SystemMessage
from pydantic import BaseModel, Field
from wbcore.contrib.ai.exceptions import APIStatusErrors
from wbcore.contrib.ai.llm.utils import run_llm

logger = logging.getLogger("llm")


class CompanyModel(BaseModel):
    name: str = Field(..., description="The name of the publicly listed company.")
    isin: str | None = Field(..., description="The ISIN of the publicly listed company. May be null if not available.")
    ticker: str | None = Field(
        ..., description="The ticker of the publicly listed company. May be null if not available."
    )
    refinitiv_identifier_code: str | None = Field(
        ..., description="The RIC of the publicly listed company. May be null if not available."
    )
    sentiment: int = Field(
        ...,
        ge=1,
        le=4,
        description="The sentiment of the news article towards the company. 1 is negative, 2 is neutral, 3 is positive, 4 is very positive.",
    )
    analysis: str = Field(
        ...,
        description="The analysis of the news article towards the company. This should include an explanation of the sentiment.",
    )


class CompaniesModel(BaseModel):
    companies: list[CompanyModel] = Field(
        default=[], description="The list of companies mentioned in the news article."
    )


@shared_task(
    autoretry_for=tuple(APIStatusErrors),
    retry_backoff=10,
    max_retries=5,  # retry i5 times maximum
    default_retry_delay=30,  # retry in 30s
    retry_jitter=False,
)
def run_company_extraction_llm(title: str, description: str, *args) -> list[dict[str, Any]]:
    from wbfdm.import_export.handlers.instrument import InstrumentLookup
    from wbfdm.models import Instrument

    relationships = []

    try:
        res = run_llm(
            prompt=[
                SystemMessage(
                    content="You will be parsed a news article, please provide the name of the publicly listed companies mentioned in the article, along with their ISIN, ticker, RIC, sentiment, and analysis."
                ),
                HumanMessage(content=f"Title: {title}, Description: {description}"),
            ],
            output_model=CompaniesModel,
        )
        instrument_ct = ContentType.objects.get_for_model(Instrument)
        for company in res.get("companies", []):
            instrument = InstrumentLookup(Instrument).lookup(
                only_security=True,
                name=company.name,
                isin=company.isin,
                ticker=company.ticker,
                refinitiv_identifier_code=company.refinitiv_identifier_code,
            )
            if instrument is not None:
                relationships.append(
                    {
                        "content_type_id": instrument_ct.id,
                        "object_id": instrument.get_root().id,
                        "sentiment": company.sentiment,
                        "analysis": company.analysis,
                        "content_object_repr": str(instrument),
                    }
                )
    except tuple(APIStatusErrors) as e:  # for APIStatusError, we let celery retry it
        raise e
    except Exception as e:  # otherwise we log the error and silently fail
        logger.warning(str(e))
    return relationships



from typing import Optional

from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbfdm.models import Instrument, InstrumentType

perf_formatting_rules = [
    dp.FormattingRule(
        style={
            "color": WBColor.RED_DARK.value,
        },
        condition=("<", 0),
    ),
    dp.FormattingRule(
        style={
            "color": WBColor.GREEN_DARK.value,
        },
        condition=(">", 0),
    ),
]
perf_formatting_rules_bold = [
    dp.FormattingRule(
        style={
            "color": WBColor.RED_DARK.value,
            "fontWeight": "bold",
        },
        condition=("<", 0),
    ),
    dp.FormattingRule(
        style={
            "color": WBColor.GREEN_DARK.value,
            "fontWeight": "bold",
        },
        condition=(">", 0),
    ),
]

instrument_formatting_rules = [
    dp.Formatting(
        column="instrument_vs_benchmark",
        formatting_rules=[
            dp.FormattingRule(
                style={"color": WBColor.GREEN_DARK.value},
                condition=("==", True),
            ),
            dp.FormattingRule(
                style={"color": WBColor.RED_DARK.value},
                condition=("==", False),
            ),
        ],
    ),
]

benchmark_formatting_rules = [
    dp.Formatting(
        column="instrument_vs_benchmark",
        formatting_rules=[
            dp.FormattingRule(
                style={"color": WBColor.GREEN_DARK.value},
                condition=("==", False),
            ),
            dp.FormattingRule(
                style={"color": WBColor.RED_DARK.value},
                condition=("==", True),
            ),
        ],
    ),
]

instrument_one_year_formatting_rules = [
    dp.Formatting(
        column="instrument_vs_benchmark_one_year",
        formatting_rules=[
            dp.FormattingRule(
                style={"color": WBColor.GREEN_DARK.value},
                condition=("==", True),
            ),
            dp.FormattingRule(
                style={"color": WBColor.RED_DARK.value},
                condition=("==", False),
            ),
        ],
    ),
]

benchmark_one_year_formatting_rules = [
    dp.Formatting(
        column="instrument_vs_benchmark_one_year",
        formatting_rules=[
            dp.FormattingRule(
                style={"color": WBColor.GREEN_DARK.value},
                condition=("==", False),
            ),
            dp.FormattingRule(
                style={"color": WBColor.RED_DARK.value},
                condition=("==", True),
            ),
        ],
    ),
]


class InstrumentPriceDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        base_fields = [
            dp.Field(key="date", label="Date"),
            dp.Field(key="net_value", label="Net Value"),
            dp.Field(key="net_value_usd", label="Net Value ($)"),
            dp.Field(key="daily_diff_net_value", label="Daily Diff.", formatting_rules=perf_formatting_rules),
            dp.Field(key="gross_value", label="Gross Value"),
            dp.Field(key="daily_diff_gross_value", label="Daily Diff.", formatting_rules=perf_formatting_rules),
            dp.Field(key="market_capitalization", label="Market Cap./AUM"),
            dp.Field(key="volume", label="Volume"),
            dp.Field(key="volume_50d", label="Avg Volume (50 days)"),
        ]
        if instrument_id := self.view.kwargs.get("instrument_id", None):
            instrument = Instrument.objects.get(id=instrument_id)
            if instrument.instrument_type == InstrumentType.PRODUCT:
                base_fields.extend(
                    [
                        dp.Field(key="sharpe_ratio", label="Sharpe Ratio"),
                        dp.Field(key="correlation", label="Correlation"),
                        dp.Field(key="beta", label="Beta"),
                        dp.Field(key="outstanding_shares_consolidated", label="Outstanding Shares"),
                    ]
                )
        return dp.ListDisplay(
            fields=base_fields,
            legends=[
                dp.Legend(
                    key="calculated",
                    items=[
                        dp.LegendItem(icon=WBColor.YELLOW.value, label="Calculated", value=True),
                        dp.LegendItem(icon=WBColor.GREEN_LIGHT.value, label="Real", value=False),
                    ],
                ),
            ],
            formatting=[
                dp.Formatting(
                    column="calculated",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW.value},
                            condition=("==", True),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", False),
                        ),
                    ],
                ),
            ],
        )


class FinancialStatisticsInstrumentPandasDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(
                    key="financial",
                    label="Financial Statistics",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={
                                "fontWeight": "bold",
                            },
                        ),
                    ],
                ),
                dp.Field(
                    key="instrument_statistics", label="Instrument", formatting_rules=instrument_formatting_rules
                ),
                dp.Field(key="benchmark_statistics", label="Benchmark", formatting_rules=benchmark_formatting_rules),
                dp.Field(
                    key="instrument_one_year",
                    label="Instrument - One Year",
                    formatting_rules=instrument_one_year_formatting_rules,
                ),
                dp.Field(
                    key="benchmark_one_year",
                    label="Benchmark - One Year",
                    formatting_rules=benchmark_one_year_formatting_rules,
                ),
            ]
        )


class BestAndWorstReturnsInstrumentPandasDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="date_best_returns", label="Date Best Returns"),
                dp.Field(key="best_returns", label="Best Returns"),
                dp.Field(key="date_worst_returns", label="Date Worst Returns"),
                dp.Field(key="worst_returns", label="Worst Returns"),
            ]
        )



import enum
from datetime import date
from types import DynamicClassAttribute

import numpy as np
import pandas as pd
from django.db import models
from wbcore.contrib.currency.models import CurrencyFXRates

from wbfdm.models import Instrument


class PeriodChoices(models.TextChoices):
    TTM = "TTM", "TTM"
    NTM = "FTM", "FTM"
    FY1 = "FY1", "FY+1"
    FY2 = "FY2", "FY+2"


class VariableChoicesChartLabel(enum.Enum):
    PE = "Price-to-Earnings PE"
    PEG = "Price-to-Earnings-to-Growth PEG"
    PS = "Price-to-Sales PS"
    PFCF = "Price-to-FreeCashflow P/FCF"
    EVEBITDA = "EV-to-EBITDA"
    EV = "Enterprise Value"
    EPSG = "EPS growth"
    REVG = "Revenue growth"
    FCFG = "FCF growth"
    MKTCAP = "Market Cap"


class VariableFormatChoices(enum.Enum):
    PE = ".1f}x"
    PEG = ".2f}"
    PS = ".1f}x"
    PFCF = ".1f}x"
    EVEBITDA = ".1f}x"
    EV = ".0f}"
    EPSG = ".1%}"
    REVG = ".1%}"
    FCFG = ".1%}"
    MKTCAP = ",.0f}bn"


class VariableChoices(models.TextChoices):
    PE = "PE", "P/E"
    PEG = "PEG", "PEG"
    PS = "PS", "P/S"
    PFCF = "PFCF", "P/FCF"
    EVEBITDA = "EVEBITDA", "EV/EBITDA"
    EV = "EV", "Enterprise Value"
    EPSG = "EPSG", "EPS Growth"
    REVG = "REVG", "Revenue Growth"
    FCFG = "FCFG", "FCF Growth"
    MKTCAP = "MKTCAP", "Market Cap"

    @DynamicClassAttribute
    def format(self):
        """The format of the Enum member."""
        return VariableFormatChoices[self._value_].value

    @DynamicClassAttribute
    def chart_label(self):
        """The chart_label of the Enum member."""
        return VariableChoicesChartLabel[self._value_].value


def _interpolate(decreasing: pd.Series, increasing: pd.Series, factor) -> float:
    return (decreasing * (1 - factor)) + (increasing * factor)


def _adapt_growth(first, second) -> np.ndarray:
    """
    In finance some variables, e.g., earnings, may pass from negative to positive and a standard percentage change
    formula is inappropriate. This formula modifies the calculation for such cases.

    Parameters
    ----------
    first = past data point/series
    second = more recent data point/series

    Returns
    -------
    A float number representing growth (percentage change)

    """
    return np.where(first > 0, second / first - 1, (second + abs(first)) / abs(first))


class FinancialAnalysisGenerator:
    def __init__(self, instruments: list, from_date: date, to_date: date, with_related: bool = False):
        """
        Initialize a generator for a specific instrument

        Parameters
        ----------
        instrument = instrument of choice
        from_date
        to_date
        with_related = extend the instrument to a list that includes all related instruments
        """
        self.instruments = instruments
        self.from_date = from_date
        self.to_date = to_date

        if with_related:
            self.instruments.extend(
                list(
                    Instrument.objects.filter(
                        models.Q(instrument_classification_related__classified_instrument__instrument__in=instruments)
                    )
                )
            )
        self.instruments_repr_map = {i.id: i.name_repr for i in self.instruments}
        self.currency_map = {i.id: i.currency.id for i in self.instruments}

    def build_df(
        self,
        instrument_prices_field_names: list[str] = [],
        fundamental_field_names: list[str] = [],
        forecast_field_names: list[str] = [],
        daily_fundamental_field_names: list[str] = [],
    ):
        """
        Used to returns a df with all the variables passed in four separate lists

        Parameters
        ----------
        instrument_prices_field_names
        fundamental_field_names
        forecast_field_names
        daily_fundamental_field_names

        Returns
        -------
        A df with all the variables in columns and date/instrument index
        """

        # df_list = []
        # if instrument_prices_field_names:
        #     df_append = pd.DataFrame(
        #         InstrumentPrice.objects.filter(
        #             instrument__in=self.instruments, date__gte=self.from_date, date__lte=self.to_date, calculated=False
        #         ).values("date", "instrument", "net_value", "outstanding_shares")
        #     )
        #     if not df_append.empty:
        #         df_list.append(df_append.set_index(["instrument", "date"]))
        #
        # end_date_df = pd.DataFrame(
        #     FiscalPeriod.objects.filter(
        #         period_type=FiscalPeriod.PeriodTypeChoice.ANNUAL,
        #         period_interim=False,
        #         instrument__in=self.instruments,
        #     ).values("period_end_date", "instrument")
        # )
        # if not end_date_df.empty:
        #     if fundamental_field_names:
        #         df_append = pd.DataFrame(
        #             Fundamental.annual_objects.filter(
        #                 instrument__in=self.instruments,
        #                 period__date_range__overlap=DateRange(self.from_date, self.to_date),
        #             ).values(
        #                 "instrument",
        #                 "period__period_end_date",
        #                 *fundamental_field_names,
        #             )
        #         ).rename(columns={"period__period_end_date": "date"})
        #         if not df_append.empty:
        #             df_list.append(df_append.set_index(["instrument", "date"]))
        #
        #     if forecast_field_names:
        #         df_append = pd.DataFrame(
        #             Forecast.objects.filter(
        #                 instrument__in=self.instruments, date__gte=self.from_date, date__lte=self.to_date
        #             ).values("instrument", "date", *forecast_field_names)
        #         )
        #         if not df_append.empty:
        #             df_list.append(df_append.set_index(["instrument", "date"]))
        #
        #     if daily_fundamental_field_names:
        #         df_append = pd.DataFrame(
        #             DailyFundamental.objects.annotate(free_cash_flow_ttm=F("free_cash_flow"))
        #             .filter(instrument__in=self.instruments, date__gte=self.from_date, date__lte=self.to_date)
        #             .values("instrument", "date", *daily_fundamental_field_names)
        #         )
        #         if not df_append.empty:
        #             df_list.append(df_append.set_index(["instrument", "date"]))
        #
        #     df = pd.concat(df_list, axis=1).sort_index().astype(float)
        #
        #     if df.columns.duplicated().any():
        #         raise ValueError("You probably have a duplicated field in the field name list")
        #
        #     if df.columns.symmetric_difference(
        #         instrument_prices_field_names
        #         + fundamental_field_names
        #         + forecast_field_names
        #         + daily_fundamental_field_names
        #     ).empty:
        #         return df.merge(
        #             end_date_df.sort_values(by="period_end_date")
        #             .groupby("instrument")
        #             .last()
        #             .rename(columns={"period_end_date": "end_date"}),
        #             left_on="instrument",
        #             right_index=True,
        #         )
        return pd.DataFrame()

    def convert_fx(self, df: pd.DataFrame, foreign_fx_field_names: list[str]) -> pd.DataFrame:
        """
        A function to FX

        Parameters
        ----------
        df = dataframe that has one or plus columns to be converted
        foreign_fx_field_names = column names to be converted

        Returns
        -------
        A dataframe with one or more columns converted with FX
        """
        df = df.set_index(["instrument", "date"])
        df["currency"] = df.index.get_level_values("instrument").map(self.currency_map)
        currencies = df["currency"].unique()
        df = df.reset_index().set_index(["date", "currency"])
        df_fx_rate = pd.DataFrame(
            CurrencyFXRates.objects.order_by("date", "currency__id")
            .filter(date__range=[self.from_date, self.to_date], currency__in=currencies)
            .values("date", "currency", "value")
        ).set_index(["date", "currency"])
        df = df.join(df_fx_rate).rename(columns={"value": "fx"}).reset_index()
        df = df.set_index(["instrument", "date"]).groupby("instrument").ffill()
        df[foreign_fx_field_names] = df[foreign_fx_field_names].div(df["fx"].astype(float), axis=0)

        return df.reset_index()

    @staticmethod
    def clean_data(
        df: pd.DataFrame,
        var_list: list[str] = [],
        drop_negative=True,
        q_low: float = 0.05,
        q_high: float = 0.95,
        z_max: int = 100,
        smooth_range: int = 3,
    ) -> pd.DataFrame:
        """
        A function to clean time-series.

        1) Drops negative values (financial ratios are non-meaningful in this case)
        2) Drops outliers above and below a certain quartile
        3) Drops values above a certain threshold

        Parameters
        ----------
        var_list = a list of variables to clean
        drop_negative = boolean to consider only positive numbers
        q_low = lower cut-off quartile
        q_high = upper cut-off quartile
        z_max = upper absolute cutoff
        smooth_range = rolling period for simple moving average smoothing

        Returns
        -------
        A clean time-series
        """
        df_temp = df.loc[:, (var_list, slice(None))].copy()
        if drop_negative:
            df_temp = df_temp.mask(df_temp < 0)
        df_temp = df_temp.where(
            df_temp.ge(df_temp.quantile(q_low), axis=1) & df_temp.le(df_temp.quantile(q_high), axis=1)
        )
        df_temp.loc[:, var_list] = df_temp[var_list].where(df_temp[var_list] < z_max)
        df_temp.ffill(inplace=True)
        df_temp.loc[:, var_list] = df_temp.loc[:, var_list].rolling(window=smooth_range).mean()
        df.loc[:, (var_list, slice(None))] = df_temp
        return df

    def get_common_valuation_ratios_df(self, period: PeriodChoices, clean_data: bool = True) -> pd.DataFrame:
        """
        Calculates financial valuation ratios

        Parameters
        ----------
        period = financial period, i.e., Next Twelve Months, Trailing Twelve Months, Fiscal Year +1 and +2
        clean_data = a boolean to apply a clean_series method

        Returns
        -------
        A dataframe with common valuation ratios for all instruments of interest

        """
        df = self.build_df(
            ["net_value", "outstanding_shares"],
            ["revenue", "net_debt", "ebitda", "eps", "free_cash_flow"],
            [
                "revenue_y1",
                "entreprise_value_y1",
                "net_debt_y1",
                "ebitda_y1",
                "eps_y1",
                "free_cash_flow_y1",
                "revenue_y2",
                "entreprise_value_y2",
                "net_debt_y2",
                "ebitda_y2",
                "eps_y2",
                "free_cash_flow_y2",
            ],
            ["eps_ttm", "eps_ftw"],
        )

        if df.empty:
            return df

        if clean_data:
            df = df.groupby(level=0).ffill().groupby(level=0).bfill()

        # Calculate a decimal factor to know how many days an instrument has till its "annual report" publication
        # so Trailing Twelve Month (TTM) and Next(or Forward) Twelve Month (NTM/FTM) interpolated variables
        # may be calculated
        df["factor"] = (365 - (df.end_date - df.index.get_level_values("date")).dt.days) % 365 / 365
        df["free_cash_flow_ntm"] = _interpolate(df["free_cash_flow_y1"], df["free_cash_flow_y2"], df["factor"])
        df["free_cash_flow_ttm"] = _interpolate(df["free_cash_flow"], df["free_cash_flow_y1"], df["factor"])
        df["revenue_ntm"] = _interpolate(df["revenue_y1"], df["revenue_y2"], df["factor"])
        df["revenue_ttm"] = _interpolate(df["revenue"], df["revenue_y1"], df["factor"])
        df["mktcap"] = (df["net_value"] * df["outstanding_shares"] / 1000000000).replace(0, np.inf)
        match period:
            case PeriodChoices.NTM:
                df["revenue"] = df["revenue_ntm"]
                df["eps"] = df["eps_ftw"]
                df["ebitda"] = _interpolate(df["ebitda_y1"], df["ebitda_y2"], df["factor"])
                df["free_cash_flow"] = _interpolate(df["free_cash_flow_y1"], df["free_cash_flow_y2"], df["factor"])
                df["net_debt"] = _interpolate(df["net_debt_y1"], df["net_debt_y2"], df["factor"])
                df["ev"] = (df["net_value"] * df["outstanding_shares"]) + df["net_debt"]
                df["epsg"] = _adapt_growth(df["eps_ttm"], df["eps_ftw"])
                df["fcfg"] = _adapt_growth(df["free_cash_flow_ttm"], df["free_cash_flow_ntm"])
                df["revg"] = _adapt_growth(df["revenue_ttm"], df["revenue_ntm"])
            case PeriodChoices.TTM:
                df["revenue"] = df["revenue_ttm"]
                df["eps"] = df["eps_ttm"]
                df["ebitda"] = _interpolate(df["ebitda"], df["ebitda_y1"], df["factor"])
                df["free_cash_flow"] = _interpolate(df["free_cash_flow"], df["free_cash_flow_y1"], df["factor"])
                df["net_debt"] = _interpolate(df["net_debt"], df["net_debt_y1"], df["factor"])
                df["ev"] = df["net_value"] * df["outstanding_shares"] + df["net_debt"]
                df["epsg"] = _adapt_growth(df["eps_ttm"].shift(250), df["eps_ttm"])
                df["fcfg"] = _adapt_growth(df["free_cash_flow_ttm"].shift(250), df["free_cash_flow_ttm"])
                df["revg"] = _adapt_growth(df["revenue"], df["revenue_ttm"])
            case default:  # noqa
                df["ev"] = df["entreprise_value_y1"]
                df["ebitda"] = df["ebitda_y1"]
                df["free_cash_flow"] = df["free_cash_flow_y1"]
                df["epsg"] = _adapt_growth(df["eps"], df["eps_y1"])
                df["eps"] = df["eps_y1"]
                df["fcfg"] = _adapt_growth(df["free_cash_flow"], df["free_cash_flow_y1"])
                df["revg"] = _adapt_growth(df["revenue"], df["revenue_y1"])
                if period == PeriodChoices.FY2:
                    df["ev"] = df["entreprise_value_y2"]
                    df["eps"] = df["eps_y2"]
                    df["ebitda"] = df["ebitda_y2"]
                    df["free_cash_flow"] = df["free_cash_flow_y2"]
                    df["epsg"] = ((1 + _adapt_growth(df["eps_y1"], df["eps_y2"])) * (1 + df["epsg"])) - 1
                    df["fcfg"] = (
                        (_adapt_growth(df["free_cash_flow_y1"], df["free_cash_flow_y2"])) * (1 + df["free_cash_flow"])
                    ) - 1
                    df["revg"] = ((_adapt_growth(df["revenue_y1"], df["revenue_y2"])) * (1 + df["revg"])) - 1

        df = df.interpolate(method="linear")

        # After getting the dataframe, calculate the ratios, then optionally clean the final series
        df["pe"] = df["net_value"] / df["eps"]
        df["peg"] = df["net_value"] / df["eps"] / (100 * df["epsg"])
        df["ps"] = df["net_value"] / (df["revenue"] / df["outstanding_shares"])
        df["pfcf"] = df["net_value"] / df["free_cash_flow"]
        df["evebitda"] = df["ev"] / df["ebitda"]

        if clean_data:
            df = df.unstack("instrument")
            df = self.clean_data(df=df, var_list=["pe", "pfcf", "evebitda"])
            df = self.clean_data(df=df, var_list=["ps"], z_max=40)
            df = self.clean_data(df=df, var_list=["peg"], z_max=15, smooth_range=10).replace(0, np.nan)
            df = self.clean_data(df=df, var_list=["ev", "epsg", "revg", "fcfg"], drop_negative=False)
            df = df.stack("instrument")

        df = df.reset_index()
        df = pd.concat([df[df.instrument == self.instruments[0].id], df[df.instrument != self.instruments[0].id]])
        df["instrument_title"] = df["instrument"].map(self.instruments_repr_map)

        df = self.convert_fx(df, ["ev", "mktcap"]).replace([np.inf, -np.inf, np.nan], None)

        return df[
            [
                "instrument",
                "instrument_title",
                "date",
                "pe",
                "peg",
                "ps",
                "pfcf",
                "evebitda",
                "ev",
                "mktcap",
                "epsg",
                "revg",
                "fcfg",
            ]
        ]

    def get_earnings_df(self, period: PeriodChoices, clean_data: bool = True) -> pd.DataFrame:
        """
        Calculates earnings anylysis

        Parameters
        ----------
        period = financial period, i.e., Next Twelve Months, Trailing Twelve Months, Fiscal Year +1 and +2
        clean_data = a boolean to apply a clean_series method

        Returns
        -------
        A dataframe with earnings analysis for all instruments of interest

        """
        df = self.build_df(
            [],
            [],
            ["eps_y1", "eps_y2"],
            ["eps_ttm", "eps_ftw"],
        )

        if df.empty:
            return df

        if clean_data:
            df = df.groupby(level=0).ffill()

        match period:
            case PeriodChoices.FY1:
                df["eps"] = df["eps_y1"]
            case PeriodChoices.FY2:
                df["eps"] = df["eps_y2"]
            case PeriodChoices.TTM:
                df["eps"] = df["eps_ttm"]
            case default:  # noqa
                df["eps"] = df["eps_ftw"]

        df = df.interpolate(method="linear")

        if clean_data:
            df = df.unstack("instrument")
            df = self.clean_data(
                df=df, var_list=["eps"], drop_negative=False, q_low=0.01, q_high=0.99, z_max=200, smooth_range=4
            )
            df = df.stack("instrument")

        df = df.reset_index()
        df = pd.concat([df[df.instrument == self.instruments[0].id], df[df.instrument != self.instruments[0].id]])
        df["instrument_title"] = df["instrument"].map(self.instruments_repr_map)

        df = self.convert_fx(df, ["eps"]).replace([np.inf, -np.inf, np.nan], None)

        return df[
            [
                "instrument",
                "instrument_title",
                "date",
                "eps",
            ]
        ]



from contextlib import suppress
from datetime import date
from typing import Generator

import pandas as pd
from django.db.models import Avg, Sum
from wbcore.serializers.fields.number import DisplayMode

from wbfdm.enums import Financial, PeriodType, SeriesType
from wbfdm.models import Instrument, InstrumentPrice

from ..decorators import register
from ..dto import Metric, MetricField, MetricKey
from ..exceptions import MetricInvalidParameterException
from .base import BaseDataloader, InstrumentMetricBaseBackend

STATISTICS_METRIC = MetricKey(
    key="statistic",
    label="Statistic",
    subfields=[
        MetricField(
            key="revenue_y_1",
            label="Revenue Y-1",
            aggregate=Sum,
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "{{currency_symbol}}"}],
            serializer_kwargs={"display_mode": DisplayMode.SHORTENED},
        ),
        MetricField(
            key="revenue_y0",
            label="Revenue Y0",
            aggregate=Sum,
            decorators=[{"position": "left", "value": "{{currency_symbol}}"}],
            serializer_kwargs={"display_mode": DisplayMode.SHORTENED},
        ),
        MetricField(
            key="revenue_y1",
            label="Revenue Y1",
            aggregate=Sum,
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "{{currency_symbol}}"}],
            serializer_kwargs={"display_mode": DisplayMode.SHORTENED},
        ),
        MetricField(
            key="market_capitalization",
            label="Market Capitalization",
            aggregate=Sum,
            decorators=[{"position": "left", "value": "{{currency_symbol}}"}],
            serializer_kwargs={"display_mode": DisplayMode.SHORTENED},
        ),
        MetricField(
            key="price",
            label="Price",
            aggregate=Avg,
            decorators=[{"position": "left", "value": "{{currency_symbol}}"}],
        ),
        MetricField(
            key="volume_50d",
            label="Volume 50D",
            aggregate=Avg,
            serializer_kwargs={"display_mode": DisplayMode.SHORTENED},
        ),
    ],
)

STATISTICS_METRIC_USD = MetricKey(
    key="statistic_usd",
    label=STATISTICS_METRIC.label,
    subfields=STATISTICS_METRIC.subfields,
    additional_prefixes=STATISTICS_METRIC.additional_prefixes,
)


class Dataloader(BaseDataloader):
    METRIC_KEY = "statistic"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.aggregate_callback = {
            "revenue_y_1": "sum",
            "revenue_y0": "sum",
            "revenue_y1": "sum",
            "market_capitalization": "sum",
            "price": "mean",
            "volume_50d": "mean",
        }

    def _compute(self) -> dict[str, float]:
        """
        Compute/fetch the statistics metrics. If the basket is constituted of multiple instrument, take the average of each performance

        Returns:
            The metrics as dictionary
        """
        if self.val_date:
            pivot_year = self.val_date.year
            instruments = self.basket_objects
            df_revenue = pd.DataFrame(
                instruments.dl.financials(
                    values=[Financial.REVENUE],
                    series_type=SeriesType.COMPLETE,
                    period_type=PeriodType.ANNUAL,
                    from_year=pivot_year - 1,
                    to_year=pivot_year + 1,
                )
            )
            if not df_revenue.empty:
                df_revenue = (
                    df_revenue.pivot_table(index="instrument_id", columns="year", values="value")
                    .rename(
                        columns={pivot_year - 1: "revenue_y_1", pivot_year: "revenue_y0", pivot_year + 1: "revenue_y1"}
                    )
                    .astype(float)
                )
            df_price = (
                pd.DataFrame(
                    InstrumentPrice.objects.filter(instrument__in=instruments, calculated=False, date=self.val_date)
                    .annotate(fx_rate=self.fx_rate_expression)
                    .annotate_market_data()
                    .values_list(
                        "instrument",
                        "internal_market_capitalization",
                        "net_value",
                        "volume_50d",
                        "fx_rate",
                    ),
                    columns=[
                        "instrument",
                        "market_capitalization",
                        "price",
                        "volume_50d",
                        "fx_rate",
                    ],
                )
                .set_index("instrument")
                .astype(float)
            )
            fx_rate = df_price["fx_rate"]
            df = pd.concat([df_revenue, df_price.drop("fx_rate", axis=1)], axis=1)
            for key in ["revenue_y_1", "revenue_y0", "revenue_y1", "market_capitalization", "price"]:
                if key in df.columns:
                    df[key] = df[key] / fx_rate

            if not df.empty:
                return (
                    df.reset_index()
                    .agg({k: v for k, v in self.aggregate_callback.items() if k in df.columns})
                    .dropna()
                    .to_dict()
                )
        return dict()


@register(move_first=True)
class InstrumentFinancialStatisticsMetricBackend(InstrumentMetricBaseBackend):
    statistic = STATISTICS_METRIC
    keys = [STATISTICS_METRIC]

    def compute_metrics(self, basket: Instrument) -> Generator[Metric, None, None]:
        val_date = self._get_valid_date(basket)
        metrics = Dataloader(basket, val_date, target_currency_key=self.TARGET_CURRENCY_KEY).compute()
        yield Metric(
            metrics=metrics,
            basket_id=basket.id,
            basket_content_type_id=self.content_type.id,
            key=self.statistic.key,
            date=None,
        )

    def _get_valid_date(self, instrument: Instrument) -> date:
        if self.val_date is None and instrument.last_valuation_date:
            return instrument.last_valuation_date
        elif self.val_date:
            with suppress(InstrumentPrice.DoesNotExist):
                return instrument.valuations.filter(date__lte=self.val_date).latest("date").date
        raise MetricInvalidParameterException()


@register(move_first=True)
class InstrumentFinancialStatisticsUSDMetricBackend(InstrumentFinancialStatisticsMetricBackend):
    statistic = STATISTICS_METRIC_USD
    keys = [STATISTICS_METRIC_USD]
    TARGET_CURRENCY_KEY = "USD"



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class InstrumentEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbfdm:instrument-list", args=[], request=self.request)

    def get_delete_endpoint(self, **kwargs):
        return None

    def get_create_endpoint(self, **kwargs):
        return reverse("wbfdm:instrumentrequest-list", args=[], request=self.request)



from contextlib import suppress
from datetime import date, timedelta
from typing import Any, Generator, Iterable

import numpy as np
import pandas as pd
from dateutil.relativedelta import relativedelta
from django.db.models import Avg, BooleanField, DateField, F, QuerySet

from wbfdm.models import Instrument, InstrumentPrice, RelatedInstrumentThroughModel

from ..decorators import register
from ..dto import Metric, MetricField, MetricKey
from .base import BaseDataloader, InstrumentMetricBaseBackend

PERFORMANCE_METRIC = MetricKey(
    key="performance",
    label="Performance",
    subfields=[
        MetricField(key="daily", label="Daily", serializer_kwargs={"percent": True, "precision": 4}, aggregate=Avg),
        MetricField(key="weekly", label="Weekly", serializer_kwargs={"percent": True, "precision": 4}, aggregate=Avg),
        MetricField(
            key="monthly", label="Monthly", serializer_kwargs={"percent": True, "precision": 4}, aggregate=Avg
        ),
        MetricField(
            key="quarterly", label="Quarterly", serializer_kwargs={"percent": True, "precision": 4}, aggregate=Avg
        ),
        MetricField(key="yearly", label="Yearly", serializer_kwargs={"percent": True, "precision": 4}, aggregate=Avg),
        MetricField(
            key="week_to_date", label="WTD", serializer_kwargs={"percent": True, "precision": 4}, aggregate=Avg
        ),
        MetricField(
            key="month_to_date", label="MTD", serializer_kwargs={"percent": True, "precision": 4}, aggregate=Avg
        ),
        MetricField(
            key="quarter_to_date", label="QTD", serializer_kwargs={"percent": True, "precision": 4}, aggregate=Avg
        ),
        MetricField(
            key="year_to_date", label="YTD", serializer_kwargs={"percent": True, "precision": 4}, aggregate=Avg
        ),
        MetricField(
            key="previous_week_to_date",
            label="Previous Week",
            serializer_kwargs={"percent": True, "precision": 4},
            aggregate=Avg,
        ),
        MetricField(
            key="previous_month_to_date",
            label="Previous Month",
            serializer_kwargs={"percent": True, "precision": 4},
            aggregate=Avg,
        ),
        MetricField(
            key="previous_quarter_to_date",
            label="Previous Quarter",
            serializer_kwargs={"percent": True, "precision": 4},
            aggregate=Avg,
        ),
        MetricField(
            key="previous_year_to_date",
            label="Previous Year",
            serializer_kwargs={"percent": True, "precision": 4},
            aggregate=Avg,
        ),
        MetricField(
            key="inception", label="Inception", serializer_kwargs={"percent": True, "precision": 4}, aggregate=Avg
        ),
    ],
    extra_subfields=[
        MetricField(
            key="is_estimated",
            label="Estimated",
            help_text="True if the performance used a estimated price",
            field_type=BooleanField,
            aggregate=None,
        ),
        MetricField(
            key="date",
            label="Performance Date",
            help_text="The date at which the performances were computed",
            field_type=DateField,
            aggregate=None,
        ),
    ],
    additional_prefixes=["benchmark", "peer"],
)

PERFORMANCE_METRIC_USD = MetricKey(
    key="performance_usd",
    label=PERFORMANCE_METRIC.label,
    subfields=PERFORMANCE_METRIC.subfields,
    additional_prefixes=PERFORMANCE_METRIC.additional_prefixes,
)


class Dataloader(BaseDataloader):
    METRIC_KEY = "performance"

    PERFORMANCE_MAP = {
        "weekly": 7,
        "monthly": 30,
        "quarterly": 120,
        "yearly": 365,
        "daily": "B",
        "week_to_date": "W-FRI",
        "month_to_date": "BME",
        "quarter_to_date": "BQE",
        "year_to_date": "BYE",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.aggregate_callback = (
            lambda df: df.min() if df.name == "date" else df.mean()
        )  # we need to not sum the "date" column otherwise pandas crashes

    @classmethod
    def get_performance_date_map(cls, pivot_date) -> dict[str, date]:
        pivot_date = pivot_date + timedelta(days=1) - pd.tseries.offsets.BDay(1)
        return {
            "weekly": (pivot_date - relativedelta(weeks=1) + timedelta(days=1) - pd.tseries.offsets.BDay(1)).date(),
            "monthly": (pivot_date - relativedelta(months=1) + timedelta(days=1) - pd.tseries.offsets.BDay(1)).date(),
            "quarterly": (
                pivot_date - relativedelta(months=3) + timedelta(days=1) - pd.tseries.offsets.BDay(1)
            ).date(),
            "yearly": (pivot_date - relativedelta(years=1) + timedelta(days=1) - pd.tseries.offsets.BDay(1)).date(),
            "daily": (pivot_date - pd.tseries.offsets.BDay(1)).date(),
            "week_to_date": (pivot_date - pd.tseries.offsets.Week(1, weekday=4)).date(),
            "month_to_date": (pivot_date - pd.tseries.offsets.BMonthEnd(1)).date(),
            "quarter_to_date": (pivot_date - pd.tseries.offsets.BQuarterEnd(1)).date(),
            "year_to_date": (pivot_date - pd.tseries.offsets.BYearEnd(1)).date(),
        }

    def get_data(self) -> Iterable[tuple[Instrument, pd.Series, pd.Series]]:
        """
        Helper method to return the instrument prices as a pandas Series

        Args:
            instrument: The instrument to get the prices from

        Returns:
            a tuple of the prices as Series and the calculated mask (as series as well)
        """

        qs = (
            InstrumentPrice.objects.filter(instrument__in=self.basket_objects, date__lte=self.val_date)
            .annotate_base_data()
            .annotate(fx_rate=self.fx_rate_expression, price=F("net_value") / F("fx_rate"))
        )
        fields = ["instrument", "date", "price", "fx_rate", "calculated"]
        instruments_map = {i.id: i for i in self.basket_objects}

        if self.min_date:
            qs = qs.filter(date__gte=self.min_date)
        else:
            qs = qs.filter(date__gte=F("instrument__inception_date"))
        recs = qs.values_list(*fields)
        df = (
            pd.DataFrame.from_records(recs, columns=fields)
            .sort_values(by="calculated")
            .groupby(["instrument", "date"])
            .agg("first")
            .sort_index()
        )
        for instrument_id, dff in df.groupby(level=0):
            dff = dff.droplevel(0)
            dff = dff.reindex(pd.date_range(dff.index.min(), dff.index.max()), method="ffill")
            dff.index = pd.to_datetime(dff.index).date
            yield instruments_map[instrument_id], dff["price"].astype(float), dff["calculated"]

    def _compute(self) -> dict[str, float]:
        """
        Compute the performance metrics for all the PERFORMANCE_MAP keys. If the basket is constituted of multiple instrument, take the average of each performance

        Returns:
            The metrics as dictionary
        """
        res = {}
        if self.val_date:
            agg_metrics = []
            is_estimated = False
            for _, prices_df, calculated_df in self.get_data():
                if not prices_df.empty and not calculated_df.empty:
                    metrics = {}
                    is_estimated = is_estimated or bool(calculated_df.iloc[-1])
                    for performance, start_date in self.get_performance_date_map(self.val_date).items():
                        with suppress(KeyError):
                            if start_price := prices_df.loc[start_date]:
                                metrics[performance] = round(prices_df.loc[self.val_date] / start_price - 1, 6)
                            previous_start_date = self.get_performance_date_map(start_date)[performance]
                            if previous_start_price := prices_df.loc[previous_start_date]:
                                metrics[f"previous_{performance}"] = round(
                                    prices_df.loc[start_date] / previous_start_price - 1, 6
                                )

                    if not prices_df.empty and prices_df.iloc[0]:
                        metrics["inception"] = round(float(prices_df.iloc[-1] / prices_df.iloc[0] - 1), 6)
                    agg_metrics.append(metrics)
            res = (
                pd.DataFrame(agg_metrics).astype(float).mean(axis=0).replace([np.inf, -np.inf, np.nan], None).to_dict()
            )
            res["is_estimated"] = is_estimated
        return res


@register(move_first=True)
class InstrumentPerformanceMetricBackend(InstrumentMetricBaseBackend):
    performance = PERFORMANCE_METRIC
    keys = [PERFORMANCE_METRIC]

    def get_related_instrument_relationships(self, basket) -> QuerySet[RelatedInstrumentThroughModel]:
        if issubclass(basket.__class__, Instrument):
            return RelatedInstrumentThroughModel.objects.filter(instrument=basket)
        return RelatedInstrumentThroughModel.objects.none()

    def compute_metrics(self, basket: Instrument) -> Generator[Metric, None, None]:
        val_date = self._get_valid_date(basket)
        metrics = Dataloader(basket, val_date, target_currency_key=self.TARGET_CURRENCY_KEY).compute()
        instrument_relationships = self.get_related_instrument_relationships(basket)
        if instrument_relationships.exists():
            for related_type in [
                RelatedInstrumentThroughModel.RelatedTypeChoices.BENCHMARK,
                RelatedInstrumentThroughModel.RelatedTypeChoices.PEER,
            ]:
                type_metrics = []
                for rel in instrument_relationships.filter(related_type=related_type):
                    type_metrics.append(
                        Dataloader(
                            rel.related_instrument, val_date, target_currency_key=self.TARGET_CURRENCY_KEY
                        ).compute()
                    )

                type_metrics = pd.DataFrame(type_metrics).mean(axis=0).round(6)
                for subfield in self.performance.subfields:
                    if (base_value := metrics.get(subfield.key)) and (type_value := type_metrics.get(subfield.key)):
                        metrics[f"{related_type.value.lower()}_{subfield.key}"] = round(base_value - type_value, 6)
        metrics["date"] = val_date
        yield Metric(
            metrics=metrics,
            basket_id=basket.id,
            basket_content_type_id=self.content_type.id,
            key=self.performance.key,
            date=None,
        )

    def get_serializer_field_attr(self, metric_field: MetricField) -> dict[str, Any]:
        attrs = super().get_serializer_field_attr(metric_field)
        pivot_date = self.val_date
        if not pivot_date:
            pivot_date = (date.today() - pd.tseries.offsets.BDay(1)).date()
        if "previous" in metric_field.key:
            pivot_date = Dataloader.get_performance_date_map(pivot_date)[metric_field.key.replace("previous_", "")]

        with suppress(KeyError):
            start_date = Dataloader.get_performance_date_map(pivot_date)[metric_field.key.replace("previous_", "")]
            attrs["help_text"] = (
                f"The {metric_field.label} performance is computed from {start_date:%Y-%m-%d} to {pivot_date:%Y-%m-%d}"
            )
        return attrs


@register(move_first=True)
class InstrumentPerformanceUSDMetricBackend(InstrumentPerformanceMetricBackend):
    performance = PERFORMANCE_METRIC_USD
    keys = [PERFORMANCE_METRIC_USD]
    TARGET_CURRENCY_KEY = "USD"



from contextlib import suppress

from rest_framework.reverse import reverse
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons import ButtonViewConfig

from wbfdm.models import Instrument


class ChildrenInstrumentModelViewConfig(ButtonViewConfig):
    SHOW_INLINE = True

    def get_custom_buttons(self) -> set:
        with suppress(Instrument.DoesNotExist):
            instrument = Instrument.objects.get(id=self.view.kwargs.get("instrument_id"))
            if instrument.parent:
                return {
                    bt.WidgetButton(
                        label="Parent",
                        endpoint=reverse("wbfdm:instrument-detail", args=[instrument.parent.id], request=self.request),
                    ),
                }
        return set()


class InstrumentButtonViewConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self) -> set:
        return {*super().get_custom_list_instance_buttons(), *self.get_custom_instance_buttons()}

    def get_custom_instance_buttons(self):
        return {
            bt.DropDownButton(
                label="Performance Analysis",
                weight=1,
                icon=WBIcon.UNFOLD.icon,
                buttons=(
                    bt.WidgetButton(
                        weight=0,
                        label="Prices",
                        icon=WBIcon.DOLLAR.icon,
                        key="prices",
                    ),
                    bt.WidgetButton(
                        weight=1,
                        key="bestandworstreturns",
                        label="Best and Worst Returns Table",
                        icon=WBIcon.INVERSE.icon,
                    ),
                    bt.WidgetButton(
                        weight=2,
                        label="Monthly Performances",
                        icon=WBIcon.TABLE.icon,
                        key="monthly_performances",
                    ),
                    bt.WidgetButton(
                        label="Financial Statistics",
                        weight=3,
                        icon=WBIcon.CHART_WATERFALL.icon,
                        key="financial-statistics",
                    ),
                    bt.DropDownButton(
                        label="Valuations",
                        weight=4,
                        icon=WBIcon.UNFOLD.icon,
                        buttons=[
                            bt.WidgetButton(
                                key="instrumentprices", label="Price and Performance", icon=WBIcon.DOLLAR.icon
                            ),
                            bt.WidgetButton(
                                key="instrumentpricestatisticchart",
                                label="Price Statistical Chart",
                                icon=WBIcon.CHART_MULTILINE.icon,
                            ),
                        ],
                    ),
                    bt.DropDownButton(
                        label="Charts",
                        weight=1000,
                        icon=WBIcon.UNFOLD.icon,
                        buttons=(
                            bt.WidgetButton(
                                label="Market Data Chart",
                                weight=0,
                                icon=WBIcon.CHART_LINE.icon,
                                key="market_data",
                            ),
                            bt.WidgetButton(
                                label="Drawdown Chart",
                                weight=1,
                                icon=WBIcon.CHART_AREA.icon,
                                key="drawdown",
                            ),
                            bt.WidgetButton(
                                label="Performance Summary",
                                weight=2,
                                icon=WBIcon.CHART_BARS_VERTICAL.icon,
                                key="performance_summary",
                            ),
                            bt.WidgetButton(
                                label="Cumulative Returns Chart",
                                weight=3,
                                icon=WBIcon.CHART_WATERFALL.icon,
                                key="cumulativereturn",
                            ),
                            bt.WidgetButton(
                                key="distributionreturnschart",
                                label="Distribution Returns Chart",
                                weight=4,
                                icon=WBIcon.CHART_BARS_HORIZONTAL.icon,
                            ),
                            bt.WidgetButton(
                                key="price_and_volume", label="Price And Volume", weight=5, icon=WBIcon.CHART_LINE.icon
                            ),
                        ),
                    ),
                ),
            ),
            bt.DropDownButton(
                label="Financial Analysis",
                icon=WBIcon.UNFOLD.icon,
                weight=2,
                buttons=[
                    bt.WidgetButton(
                        label="Income Statement",
                        weight=0,
                        icon=WBIcon.DOCUMENT_WITH_DOLLAR.icon,
                        key="income-statement",
                    ),
                    bt.WidgetButton(
                        label="Balance Sheet",
                        weight=1,
                        icon=WBIcon.ACCOUNT_BALANCE.icon,
                        key="balance-sheet",
                    ),
                    bt.WidgetButton(
                        label="Cash Flow Statement",
                        weight=2,
                        icon=WBIcon.CURRENCY_EXCHANGE.icon,
                        key="cash-flow-statement",
                    ),
                    bt.DropDownButton(
                        label="IBES",
                        weight=3,
                        icon=WBIcon.UNFOLD.icon,
                        buttons=(
                            bt.WidgetButton(
                                label="Summary",
                                weight=0,
                                icon=WBIcon.DATA_EXPLORATION.icon,
                                key="swe-summary",
                            ),
                            bt.WidgetButton(
                                label="Income Statement / Estimates",
                                weight=1,
                                icon=WBIcon.DOCUMENT_WITH_DOLLAR.icon,
                                key="swe-income-statement",
                            ),
                            bt.WidgetButton(
                                label="Balance Sheet / Estimates",
                                weight=2,
                                icon=WBIcon.ACCOUNT_BALANCE.icon,
                                key="swe-balance-sheet",
                            ),
                            bt.WidgetButton(
                                label="Cash Flow / Estimates",
                                weight=3,
                                icon=WBIcon.CURRENCY_EXCHANGE.icon,
                                key="swe-cashflow-statement",
                            ),
                            bt.WidgetButton(
                                label="Ratios / Estimates",
                                weight=4,
                                icon=WBIcon.DATA_EXPLORATION.icon,
                                key="swe-ratios",
                            ),
                            bt.WidgetButton(
                                label="Margins",
                                weight=5,
                                icon=WBIcon.DATA_EXPLORATION.icon,
                                key="swe-margins",
                            ),
                            bt.WidgetButton(
                                label="Cashflow Ratios",
                                weight=6,
                                icon=WBIcon.DATA_EXPLORATION.icon,
                                key="swe-cashflow-ratios",
                            ),
                            bt.WidgetButton(
                                label="Asset Turnover Ratios",
                                weight=7,
                                icon=WBIcon.DATA_EXPLORATION.icon,
                                key="swe-asset-turnover-ratios",
                            ),
                            bt.WidgetButton(
                                label="Credit Ratios",
                                weight=8,
                                icon=WBIcon.DATA_EXPLORATION.icon,
                                key="swe-credit",
                            ),
                            bt.WidgetButton(
                                label="Long-Term Solvency Ratios",
                                weight=9,
                                icon=WBIcon.DATA_EXPLORATION.icon,
                                key="swe-long-term-solvency",
                            ),
                            bt.WidgetButton(
                                label="Short-Term Liquidity Ratios",
                                weight=10,
                                icon=WBIcon.DATA_EXPLORATION.icon,
                                key="swe-short-term-liquidity",
                            ),
                        ),
                    ),
                ],
            ),
            bt.WidgetButton(
                label="Officers",
                icon=WBIcon.PEOPLE.icon,
                key="officers",
            ),
            bt.DropDownButton(
                label="ESG",
                icon=WBIcon.UNFOLD.icon,
                weight=3,
                buttons=[
                    bt.WidgetButton(
                        label="Controversies",
                        icon=WBIcon.EVENT.icon,
                        key="controversies",
                    ),
                    bt.WidgetButton(
                        label="PAI",
                        icon=WBIcon.CHART_SWITCHES.icon,
                        key="pai",
                    ),
                ],
            ),
            # bt.DropDownButton(
            #     label="Financial Analysis (Old)",
            #     weight=3,
            #     icon=WBIcon.UNFOLD.icon,
            #     buttons=(
            #         bt.WidgetButton(
            #             key="summary_table",
            #             label="Summary Table",
            #             icon=WBIcon.TABLE.icon,
            #         ),
            #         bt.WidgetButton(
            #             key="financials_graph",
            #             label="Financials Graph",
            #             icon=WBIcon.CHART_BARS_HORIZONTAL.icon,
            #         ),
            #         bt.WidgetButton(
            #             key="profitability_ratios",
            #             label="Profitability Ratios",
            #             icon=WBIcon.CHART_LINE.icon,
            #         ),
            #         bt.WidgetButton(
            #             key="cash_flow_analysis_table",
            #             label="Cash Flow Analysis Table",
            #             icon=WBIcon.TABLE.icon,
            #         ),
            #         bt.WidgetButton(
            #             key="cash_flow_analysis_chart",
            #             label="Cash Flow Analysis Chart",
            #             icon=WBIcon.CHART_BARS_HORIZONTAL.icon,
            #         ),
            #         bt.WidgetButton(
            #             key="net_debt_and_ebitda_chart",
            #             label="Net Debt And EBITDA Analysis Chart",
            #             icon=WBIcon.CHART_BARS_HORIZONTAL.icon,
            #         ),
            #         bt.WidgetButton(
            #             key="valuation_ratios-old",
            #             label="Valuation Ratios",
            #             icon=WBIcon.CHART_MULTILINE.icon,
            #         ),
            #         bt.WidgetButton(
            #             key="earnings_chart",
            #             label="Earnings Analysis",
            #             icon=WBIcon.CHART_LINE.icon,
            #         ),
            #     ),
            # ),
        }



from datetime import date

import pytest
from faker import Faker
from rest_framework.test import APIRequestFactory
from wbcore import viewsets
from wbcore.utils.strings import get_aggregate_symbol

from wbfdm.contrib.metric.backends.performances import (
    PERFORMANCE_METRIC,
    InstrumentPerformanceMetricBackend,
)
from wbfdm.contrib.metric.models import InstrumentMetric
from wbfdm.models import Instrument
from wbfdm.serializers import InstrumentModelSerializer

from ..viewsets.mixins import InstrumentMetricMixin

fake = Faker()


@pytest.mark.django_db
class TestInstrumentMetricMixin:
    @pytest.fixture
    def viewset(self, instrument_metric_factory):
        metric = instrument_metric_factory.create(key="performance")

        class InstrumentWithMetricModelViewSet(InstrumentMetricMixin, viewsets.ModelViewSet):
            METRIC_KEYS = (PERFORMANCE_METRIC,)
            METRIC_WITH_PREFIXED_KEYS = fake.boolean()
            serializer_class = InstrumentModelSerializer
            queryset = Instrument.objects.all()

            @property
            def metric_basket(self):
                return metric.basket

            @property
            def metric_date(self) -> date | None:
                return metric.date

        request = APIRequestFactory().get("")
        return InstrumentWithMetricModelViewSet(request=request, format_kwarg={})

    def test_metric_serializer_fields(self, weekday, viewset):
        assert (
            viewset._metric_serializer_fields.keys()
            == InstrumentPerformanceMetricBackend(weekday)
            .get_serializer_fields(with_prefixed_key=viewset.METRIC_WITH_PREFIXED_KEYS)
            .keys()
        )

    def test_get_ordering_fields(self, viewset):
        ordering_fields = viewset.get_ordering_fields()
        for key, _ in PERFORMANCE_METRIC.get_fields(with_prefixed_key=viewset.METRIC_WITH_PREFIXED_KEYS):
            assert key in ordering_fields

    def test_get_queryset(self, viewset):
        queryset = viewset.get_queryset()
        annotations = list(queryset.query.annotations.keys())
        for key, _ in PERFORMANCE_METRIC.get_fields(with_prefixed_key=viewset.METRIC_WITH_PREFIXED_KEYS):
            assert key in annotations

    def test_get_serializer(self, viewset):
        serializer = viewset.get_serializer()
        for key, _ in PERFORMANCE_METRIC.get_fields(with_prefixed_key=viewset.METRIC_WITH_PREFIXED_KEYS):
            assert key in serializer.fields

    def test_get_aggregates(self, viewset):
        queryset = viewset.get_queryset()
        aggregates = viewset.get_aggregates(queryset, queryset)
        metric = InstrumentMetric.objects.first()
        assert metric is not None
        for key, _ in PERFORMANCE_METRIC.get_fields(with_prefixed_key=viewset.METRIC_WITH_PREFIXED_KEYS):
            key_filter = PERFORMANCE_METRIC.subfields_filter_map[key]
            subfield = PERFORMANCE_METRIC.subfields_map[key]
            assert subfield.aggregate is not None

            if key_filter in metric.metrics:
                assert aggregates[key] == {get_aggregate_symbol(subfield.aggregate.name): metric.metrics[key_filter]}



from wbcore import serializers as wb_serializers

from wbfdm.models import InstrumentPrice
from wbfdm.serializers.instruments.instruments import (
    InvestableInstrumentRepresentationSerializer,
)


class InstrumentPriceModelSerializer(wb_serializers.ModelSerializer):
    _instrument = InvestableInstrumentRepresentationSerializer(source="instrument")
    currency_symbol = wb_serializers.CharField(read_only=True)
    net_value = wb_serializers.DecimalField(max_digits=16, decimal_places=2, default=0)
    net_value_usd = wb_serializers.DecimalField(max_digits=16, decimal_places=2, default=0)
    gross_value = wb_serializers.DecimalField(max_digits=16, decimal_places=2, default=0)
    daily_diff_net_value = wb_serializers.FloatField(required=False, read_only=True, default=0, precision=4)
    daily_diff_gross_value = wb_serializers.FloatField(required=False, read_only=True, default=0, precision=4)
    real_price_exists = wb_serializers.BooleanField(default=False, read_only=True)

    class Meta:
        model = InstrumentPrice
        percent_fields = [
            "daily_diff_net_value",
            "daily_diff_gross_value",
        ]
        decorators = {
            "market_capitalization": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{currency_symbol}}"
            ),
        }
        fields = (
            "id",
            "date",
            "net_value",
            "net_value_usd",
            "gross_value",
            "calculated",
            "real_price_exists",
            "sharpe_ratio",
            "correlation",
            "beta",
            "daily_diff_net_value",
            "daily_diff_gross_value",
            "instrument",
            "_instrument",
            "volume",
            "volume_50d",
            "volume_200d",
            "currency_symbol",
            "outstanding_shares_consolidated",
            "market_capitalization",
        )


class InstrumentPriceInstrumentModelSerializer(InstrumentPriceModelSerializer):
    class Meta(InstrumentPriceModelSerializer.Meta):
        fields = (
            "id",
            "date",
            "net_value",
            "net_value_usd",
            "gross_value",
            "sharpe_ratio",
            "calculated",
            "correlation",
            "beta",
            "daily_diff_net_value",
            "daily_diff_gross_value",
            "volume",
            "volume_50d",
            "volume_200d",
            "currency_symbol",
            "outstanding_shares_consolidated",
            "market_capitalization",
        )



# Generated by Django 4.2.11 on 2024-04-10 14:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0017_instrument_instrument_security_index"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="instrument",
            index=models.Index(fields=["level"], name="instrument_level_index"),
        ),
        migrations.AddIndex(
            model_name="instrument",
            index=models.Index(fields=["parent"], name="instrument_parent_index"),
        ),
    ]



from typing import Any, Optional

from rest_framework.reverse import reverse
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display import Display
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbfdm.contrib.metric.dto import MetricField


class InstrumentMetricDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="key", label="Key"),
                dp.Field(key="date", label="Date"),
                dp.Field(key="instrument", label="Instrument"),
                dp.Field(key="parent_metric", label="Parent Metric"),
            ],
            tree=True,
            tree_group_pinned="left",
            tree_group_field="basket_repr",
            tree_group_label="Basket",
            tree_group_level_options=[
                dp.TreeGroupLevelOption(
                    filter_key="parent_metric",
                    filter_depth=1,
                    # lookup="id_repr",
                    clear_filter=True,
                    list_endpoint=reverse(
                        "metric:instrumentmetric-list",
                        args=[],
                        request=self.request,
                    ),
                )
            ],
        )

    def get_instance_display(self) -> Display:
        children_metrics_section = create_simple_section(
            "children_metrics_section", "Child Metrics", [["children_metrics"]], "children_metrics", collapsed=True
        )
        return create_simple_display(
            [
                ["basket_content_type", "basket_id", "basket_repr"],
                ["date", "key", "instrument"],
                [repeat_field(3, "metrics")],
                [repeat_field(3, "children_metrics")],
            ],
            [children_metrics_section],
        )


class InstrumentMetricPivotedListDisplayConfig(DisplayViewConfig):
    """
    Instrument metric Class to register automatically metrics fields into the list display.

    The view attribute is expected to inherit from InstrumentMetricMixin
    """

    def _get_metric_field_attr(self, field: MetricField) -> dict[str, Any]:
        attrs = dict(width=75)
        attrs.update(field.list_display_kwargs)
        return attrs

    def _get_metric_list_display(self) -> dp.ListDisplay:
        metrics = []

        for metric_key in self.view.metric_keys:
            metric_label = metric_key.label
            fields = []
            for subfield in metric_key.subfields:
                field_key = f"{metric_key.key}_{subfield.key}"
                fields.append(dp.Field(key=field_key, label=subfield.label, **self._get_metric_field_attr(subfield)))
            metrics.append(dp.Field(key=None, label=metric_label, children=fields))

        return dp.ListDisplay(fields=[dp.Field(label="Metrics", open_by_default=False, key=None, children=metrics)])

    def _get_metadata(self) -> Any:
        display = self.get_metadata()
        if "list" in display:
            list_display = self._get_metric_list_display()
            if isinstance(list_display, Display):
                metrics_list_display = list_display.serialize()
            else:
                metrics_list_display = dict(list_display or {})
            display["list"]["fields"].extend(metrics_list_display["fields"])
        return display



import calendar
import math
from collections import defaultdict
from contextlib import suppress
from datetime import date, timedelta
from decimal import Decimal
from typing import Dict, Iterable, Optional

import numpy as np
import pandas as pd
from pandas.tseries.offsets import BDay
from wbcore.contrib.currency.models import CurrencyFXRates

from wbfdm.analysis.financial_analysis.financial_statistics_analysis import (
    FinancialStatistics,
)
from wbfdm.backends.dto import PriceDTO
from wbfdm.enums import MarketData
from wbfdm.models.instruments.instrument_prices import InstrumentPrice


class InstrumentPMSMixin:
    def get_prices_df_with_calculated(self, market_data: MarketData = MarketData.CLOSE, **kwargs) -> pd.DataFrame:
        prices = pd.DataFrame(self.get_prices(values=[market_data], **kwargs)).rename(
            columns={"valuation_date": "date"}
        )
        if "calculated" not in prices.columns:
            prices["calculated"] = False

        if not prices.empty and market_data.value in prices.columns:
            prices = prices[[market_data.value, "calculated", "date"]].sort_values(by="calculated")
            prices = prices.groupby("date").first()
            prices.index = pd.to_datetime(prices.index)
            prices = prices.replace([np.inf, -np.inf, np.nan], None)
            return prices.sort_index()
        return pd.DataFrame()

    def get_prices_df(self, market_data: MarketData = MarketData.CLOSE, **kwargs) -> pd.Series:
        prices = self.get_prices_df_with_calculated(market_data=market_data, **kwargs)

        if market_data.value in prices.columns:
            return prices[market_data.value].astype(float)
        return pd.Series(dtype="float64")

    def get_price(self, val_date: date, price_date_timedelta: int = 3) -> Decimal:
        if self.is_cash:
            return Decimal(1)
        return self._build_dto(val_date, price_date_timedelta=price_date_timedelta).close

    def _build_dto(self, val_date: date, price_date_timedelta: int = 3) -> PriceDTO:  # for backward compatibility
        try:
            price = self.valuations.get(date=val_date)
            close = float(price.net_value)
            return PriceDTO(
                pk=price.id,
                instrument=self.id,
                date=val_date,
                open=close,
                close=close,
                high=close,
                low=close,
                volume=close,
                market_capitalization=price.market_capitalization,
                outstanding_shares=float(price.outstanding_shares) if price.outstanding_shares else None,
            )
        except InstrumentPrice.DoesNotExist:
            prices = sorted(
                self.get_prices(from_date=(val_date - BDay(price_date_timedelta)).date(), to_date=val_date),
                key=lambda x: x["valuation_date"],
                reverse=True,
            )
            if (
                prices
                and (p := prices[0])
                and (close := p.get("close", None))
                and (p_date := p.get("valuation_date", None))
            ):
                return PriceDTO(
                    pk=p["id"],
                    instrument=self.id,
                    date=p_date,
                    open=p.get("open", None),
                    close=close,
                    high=p.get("high", None),
                    low=p.get("low", None),
                    volume=p.get("volume", None),
                    market_capitalization=p.get("market_capitalization", None),
                    outstanding_shares=p.get("outstanding_shares", None),
                )
            raise ValueError("Not price was found")

    # Instrument Prices Utility Functions
    @classmethod
    def _compute_performance(cls, prices: pd.Series, freq: str = "BME") -> pd.DataFrame:
        if prices.empty:
            raise ValueError("Price series cannot be empty")
        performance = FinancialStatistics(prices).compute_performance(freq=freq)  # For backward compatibility
        return pd.concat([prices, performance], axis=1).dropna(
            how="any", subset=["performance"]
        )  # For backward compatibility

    @classmethod
    def extract_monthly_performance_df(cls, prices: pd.Series) -> pd.DataFrame:
        if prices.empty:
            raise ValueError("Price series cannot be empty")
        performance = FinancialStatistics(prices).extract_monthly_performance_df()  # For backward compatibility
        df = pd.concat([performance], axis=1, keys=["performance"])
        df["year"] = df.index.year
        df["month"] = df.index.month
        return df.dropna(how="any", subset=["performance"]).reset_index(drop=True)  # For backward compatibility

    @classmethod
    def extract_annual_performance_df(cls, prices: pd.Series) -> pd.DataFrame:
        if prices.empty:
            raise ValueError("Price series cannot be empty")
        performance = FinancialStatistics(prices).extract_annual_performance_df()  # For backward compatibility
        df = pd.concat([performance], axis=1, keys=["performance"])
        df["year"] = df.index.year
        return df.dropna(how="any", subset=["performance"]).reset_index(drop=True)  # For backward compatibility

    @classmethod
    def extract_inception_performance_df(cls, prices: pd.Series) -> float:
        if prices.empty:
            raise ValueError("Price series cannot be empty")
        return FinancialStatistics(prices).extract_inception_performance_df()  # For backward compatibility

    @classmethod
    def extract_daily_performance_df(cls, prices: pd.Series) -> pd.DataFrame:
        if prices.empty:
            raise ValueError("Price series cannot be empty")
        performance = FinancialStatistics(prices).extract_daily_performance_df()
        df = pd.concat([performance], axis=1, keys=["performance"])
        df["year"] = df.index.year
        return df.dropna(how="any", subset=["performance"])  # For backward compatibility

    def get_monthly_return_summary(
        self, start: Optional[date] = None, end: Optional[date] = None, **kwargs
    ) -> pd.DataFrame:
        if not (prices := self.get_prices_df_with_calculated(from_date=start, to_date=end, **kwargs)).empty:
            calculated_mask = prices[["calculated"]].copy().groupby([prices.index.year, prices.index.month]).tail(1)
            calculated_mask["year"] = calculated_mask.index.year
            calculated_mask["month"] = calculated_mask.index.month
            calculated_mask = (
                calculated_mask[["year", "month", "calculated"]]
                .reset_index(drop=True)
                .groupby(["year", "month"])
                .any()
            )
            monthly_perfs = self.extract_monthly_performance_df(prices.close)
            annual_perfs = self.extract_annual_performance_df(prices.close)
            annual_perfs["month"] = "annual"
            perfs = pd.concat([monthly_perfs, annual_perfs], axis=0, ignore_index=True)

            return perfs.replace([np.inf, -np.inf, np.nan], None), calculated_mask
        return pd.DataFrame(), pd.DataFrame()

    def get_monthly_return_summary_dict(
        self, start: Optional[date] = None, end: Optional[date] = None, **kwargs
    ) -> Dict:
        perfs, calculated_mask = self.get_monthly_return_summary(start, end, **kwargs)
        res = defaultdict(dict)
        if not perfs.empty:
            for year, df in perfs.sort_values(by="year", ascending=False).groupby("year", sort=False):
                df = df.set_index("month")
                for i in range(1, 13):
                    try:
                        perf = float(df.loc[i, "performance"])
                    except (IndexError, KeyError):
                        perf = None
                    try:
                        calculated = bool(calculated_mask.loc[(year, i), "calculated"])
                    except (IndexError, KeyError):
                        calculated = False
                    res[year][calendar.month_abbr[i]] = {"performance": perf, "calculated": calculated}
                try:
                    res[year]["annual"] = {
                        "performance": float(df.loc["annual", "performance"]),
                        "calculated": bool(calculated_mask.loc[(year, slice(None)), "calculated"].any()),
                    }
                except (IndexError, KeyError):
                    res[year]["annual"] = {"performance": None, "calculated": False}

        return dict(res)

    def build_benchmark_df(self, end_date: Optional[date] = None, **kwargs) -> pd.Series:
        df = pd.Series(dtype="float64")
        prices_df = self.get_prices_df(to_date=end_date).rename("net_value")
        if not prices_df.empty and (benchmark := self.primary_benchmark) and self.primary_risk_instrument:
            start_date = prices_df.index[0]
            end_date = prices_df.index[-1]
            kwargs = {"from_date": start_date, "to_date": end_date}
            # Get and prepare Risk free rate dataframe from stainly
            risk_df = self.primary_risk_instrument.get_prices_df(**kwargs).rename("rate")

            benchmark_df = benchmark.get_prices_df(**kwargs).rename("benchmark_net_value")
            # Prepare final dataframe, fill the NAN with backward index
            df = pd.concat([risk_df, benchmark_df, prices_df], axis=1).astype("float64").ffill(axis=0).sort_index()
            df.index = pd.to_datetime(df.index)

        return df

    def _get_price_objects(self, from_date: date, to_date: date, clear: bool = False) -> Iterable[InstrumentPrice]:
        df = pd.DataFrame(
            self.__class__.objects.filter(id=self.id).dl.market_data(
                from_date=from_date
                - timedelta(
                    days=90
                ),  # we make sure to at least import the last 80 days to be sure to be able to compute the volume 50d
                to_date=to_date,
                values=[MarketData.CLOSE, MarketData.VOLUME, MarketData.MARKET_CAPITALIZATION],
            )
        )
        if not df.empty:
            df["calculated"] = False
            df = df.set_index("valuation_date").sort_index()

            # # if market cap is not found, maybe we have better chance on the primary exhange
            if not df.market_capitalization.notnull().any() and self.parent and (company := self.parent.get_root()):
                with suppress(KeyError):
                    df["market_capitalization"] = pd.DataFrame(
                        self.__class__.objects.filter(id=company.id).dl.market_data(
                            from_date=from_date,
                            to_date=to_date,
                            values=[MarketData.MARKET_CAPITALIZATION],
                        )
                    ).set_index("valuation_date")["market_capitalization"]

            ts = pd.date_range(df.index.min(), df.index.max(), freq="B")
            # fill forward missing data
            df = df.reindex(ts)
            df[["close", "market_capitalization"]] = df[["close", "market_capitalization"]].astype(float).ffill()
            df.volume = df.volume.astype(float).fillna(0)
            df.calculated = df.calculated.astype(bool).fillna(
                True
            )  # we do not ffill calculated but set the to True to mark them as "estimated"/"not real"
            df["volume_50d"] = df["volume"].rolling(50).mean()
            df = df[df.index.date >= from_date]  # we import only from the requested from_date
            df = df.reset_index().dropna(subset=["index", "close"])
            df = df.replace([np.inf, -np.inf, np.nan], None)

            for row in df.to_dict("records"):
                if (_date := row.get("index")) and (close := row.get("close", None)):
                    # we make sure data does not exist 10 digits (for db constraint)
                    if int(math.log10(close)) + 1 < 10:
                        try:
                            try:
                                p = InstrumentPrice.objects.get(instrument=self, date=_date, calculated=False)
                            except InstrumentPrice.DoesNotExist:
                                p = InstrumentPrice.objects.get(instrument=self, date=_date, calculated=True)

                            # update only if net value is different with existing instrument price
                            if (
                                round(p.net_value, 2) != round(close, 2)
                                or (p.market_capitalization != row.get("market_capitalization"))
                                or (p.volume != row.get("volume"))
                                or (p.calculated != row.get("calculated"))
                                or clear
                            ):
                                p.net_value = close
                                p.gross_value = close
                                p.calculated = row["calculated"]
                                p.volume = row.get("volume", p.volume)
                                p.volume_50d = row.get("volume_50d", p.volume_50d)
                                p.market_capitalization = row.get("market_capitalization", p.market_capitalization)
                                p.market_capitalization_consolidated = p.market_capitalization
                                p.set_dynamic_field(False)
                                p.id = None
                                yield p
                        except InstrumentPrice.DoesNotExist:
                            with suppress(CurrencyFXRates.DoesNotExist):
                                p = InstrumentPrice(
                                    currency_fx_rate_to_usd=CurrencyFXRates.objects.get(  # we need to get the currency rate because we bulk create the object, and thus save is not called
                                        date=_date, currency=self.currency
                                    ),
                                    instrument=self,
                                    date=_date,
                                    calculated=row["calculated"],
                                    net_value=close,
                                    gross_value=close,
                                    volume=row.get("volume", None),
                                    market_capitalization=row.get("market_capitalization", None),
                                    volume_50d=row.get("volume_50d", None),
                                )
                                p.set_dynamic_field(False)
                                yield p

    @classmethod
    def bulk_save_instrument_prices(cls, objs):
        InstrumentPrice.objects.bulk_create(
            objs,
            unique_fields=["instrument", "calculated", "date"],
            update_conflicts=True,
            update_fields=[
                "net_value",
                "gross_value",
                "volume",
                "market_capitalization",
                "market_capitalization_consolidated",
                "calculated",
                "volume_50d",
            ],
            batch_size=1000,
        )



from django.apps import AppConfig
from django.utils.module_loading import autodiscover_modules


class MetricConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "wbfdm.contrib.metric"

    def ready(self) -> None:
        autodiscover_modules("metric.backends")



from import_export.widgets import ManyToManyWidget

from wbfdm.models import Classification, ClassificationGroup
from wbfdm.preferences import get_default_classification_group


class ClassificationManyToManyWidget(ManyToManyWidget):
    model = Classification
    field = "code_aggregated"

    def __init__(self, separator=None, field="code_aggregated", primary_classification_group=False, **kwargs):
        super().__init__(Classification, separator=separator, field=field, **kwargs)
        self.primary_classification_group = primary_classification_group

    def clean(self, value, row=None, **kwargs):
        if self.primary_classification_group:
            primary_classification_group = ClassificationGroup.objects.get(is_primary=True)
        else:
            primary_classification_group = get_default_classification_group()
        queryset = Classification.objects.filter(
            **{f"{self.field}__icontains": value, "group": primary_classification_group}
        )
        if queryset.count() == 1:
            return queryset



from wbcore.metadata.configs.titles import TitleViewConfig


class ValuationRatioChartTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return f"{str(self.view.instrument)} - Financial Statistics Chart"



from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class ExchangeButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return set()

    def get_custom_instance_buttons(self):
        return set()



from reversion.views import RevisionMixin
from wbcore import viewsets
from wbcore.permissions.permissions import InternalUserPermissionMixin

from wbfdm.models.instruments.instrument_lists import (
    InstrumentList,
    InstrumentListThroughModel,
)
from wbfdm.serializers.instruments.instrument_lists import (
    InstrumentListModelSerializer,
    InstrumentListRepresentationSerializer,
    InstrumentListThroughModelSerializer,
)

from ..configs.display.instrument_lists import (
    InstrumentListDisplayConfig,
    InstrumentListThroughModelDisplayConfig,
)
from ..configs.endpoints.instrument_lists import (
    InstrumentListThroughModelEndpointConfig,
    InstrumentListThroughModelInstrumentEndpointConfig,
    InstrumentListThroughModelInstrumentListEndpointConfig,
)
from ..mixins import InstrumentMixin


class InstrumentListRepresentationModelViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    search_fields = ["name"]
    queryset = InstrumentList.objects.all()
    serializer_class = InstrumentListRepresentationSerializer


class InstrumentListModelViewSet(InternalUserPermissionMixin, RevisionMixin, viewsets.ModelViewSet):
    queryset = InstrumentList.objects.all()
    serializer_class = InstrumentListModelSerializer

    display_config_class = InstrumentListDisplayConfig
    ordering = ("name",)


class InstrumentListThroughModelViewSet(InternalUserPermissionMixin, RevisionMixin, viewsets.ModelViewSet):
    queryset = InstrumentListThroughModel.objects.select_related("instrument", "instrument_list")
    serializer_class = InstrumentListThroughModelSerializer
    endpoint_config_class = InstrumentListThroughModelEndpointConfig

    filterset_fields = {
        "instrument_list": ["exact"],
        "instrument": ["exact"],
        "from_date": ["gte", "lte"],
        "to_date": ["gte", "lte"],
        "validated": ["exact"],
    }
    display_config_class = InstrumentListThroughModelDisplayConfig
    search_fields = ["instrument__name_repr", "instrument_str"]
    ordering_fields = [
        "instrument_list__name",
        "instrument__name",
        "instrument_str",
        "from_date",
        "to_date",
        "validated",
    ]


class InstrumentListThroughModelInstrumentListViewSet(InstrumentListThroughModelViewSet):
    endpoint_config_class = InstrumentListThroughModelInstrumentListEndpointConfig
    ordering = ["instrument__name"]

    def get_queryset(self):
        return super().get_queryset().filter(instrument_list=self.kwargs["instrument_list_id"])


class InstrumentListThroughModelInstrumentViewSet(InstrumentMixin, InstrumentListThroughModelViewSet):
    endpoint_config_class = InstrumentListThroughModelInstrumentEndpointConfig
    ordering = ["instrument_list__name"]

    def get_queryset(self):
        return super().get_queryset().filter(instrument__in=self.instrument.get_family())



# Generated by Django 5.0.6 on 2024-05-17 13:07

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0020_optionaggregate_option_and_more"),
    ]

    operations = [
        migrations.DeleteModel(
            name="InstrumentDailyStatistics",
        ),
    ]



import factory
import rstr
from slugify import slugify

from wbfdm.models import Instrument, InstrumentType


class InstrumentTypeFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("word")
    key = factory.LazyAttribute(lambda o: slugify(o.name))

    class Meta:
        django_get_or_create = ("key",)
        model = InstrumentType


class InstrumentFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("company")
    name_repr = factory.LazyAttribute(lambda x: x.name)
    description = factory.Faker("paragraph")
    instrument_type = factory.SubFactory(InstrumentTypeFactory)
    inception_date = factory.Faker("past_date")
    delisted_date = None
    currency = factory.SubFactory("wbcore.contrib.currency.factories.CurrencyFactory")
    country = factory.SubFactory("wbcore.contrib.geography.factories.CountryFactory")
    exchange = factory.SubFactory("wbfdm.factories.exchanges.ExchangeFactory")

    identifier = factory.LazyAttribute(lambda o: rstr.xeger("([A-Z]{2}[A-Z0-9]{9}[0-9]{1})"))
    isin = factory.LazyAttribute(lambda o: rstr.xeger("([A-Z]{2}[A-Z0-9]{9}[0-9]{1})"))
    ticker = factory.LazyAttribute(lambda o: rstr.xeger("([A-Z]{4})"))
    dl_parameters = {"market_data": {"path": "wbfdm.contrib.internal.dataloaders.market_data.MarketDataDataloader"}}

    @factory.post_generation
    def classifications(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for classification in extracted:
                self.classifications.add(classification)

    @factory.post_generation
    def related_instruments(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for instrument in extracted:
                self.related_instruments.add(instrument)

    class Meta:
        model = Instrument


class CashFactory(InstrumentFactory):
    is_cash = True
    instrument_type = factory.LazyAttribute(lambda o: InstrumentTypeFactory.create(name="Cash", key="cash"))


class EquityFactory(InstrumentFactory):
    instrument_type = factory.LazyAttribute(lambda o: InstrumentTypeFactory.create(name="Equity", key="equity"))

    class Meta:
        model = Instrument



from datetime import datetime
from io import BytesIO
from typing import Optional

from django.db import models
from pandas.tseries.offsets import BDay
from wbcore.contrib.io.backends import AbstractDataBackend, register

from .mixin import DataBackendMixin
from .utils import Controller

DEFAULT_MAPPING = {"DWFC": "free_cash", "EPS": "eps_ttm", "EPS1FD12": "eps_ftw"}


@register("Daily Fundamental", provider_key="refinitiv", save_data_in_import_source=False, passive_only=False)
class DataBackend(DataBackendMixin, AbstractDataBackend):
    def __init__(self, import_credential: Optional[models.Model] = None, **kwargs):
        self.controller = Controller(import_credential.username, import_credential.password)

    def get_files(
        self,
        execution_time: datetime,
        obj_external_ids: list[str] = None,
        **kwargs,
    ) -> BytesIO:
        execution_date = execution_time.date()
        start = kwargs.get("start", (execution_date - BDay(1)).date())

        fields = list(DEFAULT_MAPPING.keys())
        if obj_external_ids:
            df = self.controller.get_data(obj_external_ids, fields, start, execution_date)
            if not df.empty:
                content_file = BytesIO()
                df.to_json(content_file, orient="records")
                file_name = f"daily_fundamental_{start:%Y-%m-%d}-{execution_date:%Y-%m-%d}_{datetime.timestamp(execution_time)}.json"
                yield file_name, content_file
